"""Provide utils for Python 2 plugins."""
import copy
import sys
import threading
import time
import traceback
import uuid

try:
    import queue
except ImportError:
    import Queue as queue


def debounce(s):
    """Decorate to ensure function can only be called once every `s` seconds."""

    def decorate(f):
        d = {"t": None}

        def wrapped(*args, **kwargs):
            if d["t"] is None or time.time() - d["t"] >= s:
                result = f(*args, **kwargs)
                d["t"] = time.time()
                return result

        return wrapped

    return decorate


def setInterval(interval):
    """Set interval."""

    def decorator(function):
        def wrapper(*args, **kwargs):
            stopped = threading.Event()

            def loop():  # executed in another thread
                while not stopped.wait(interval):  # until stopped
                    function(*args, **kwargs)

            t = threading.Thread(target=loop)
            t.daemon = True  # stop if the program exits
            t.start()
            return stopped

        return wrapper

    return decorator


class dotdict(dict):
    """Access dictionary attributes with dot.notation."""

    __getattr__ = dict.get
    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __deepcopy__(self, memo=None):
        """Make a deep copy."""
        return dotdict(copy.deepcopy(dict(self), memo=memo))


def getKeyByValue(d, value):
    """Return key by value."""
    for k, v in d.items():
        if value == v:
            return k
    return None


class ReferenceStore:
    """Represent a reference store."""

    def __init__(self):
        """Set up store."""
        self._store = {}

    def _genId(self):
        """Generate an id."""
        return str(uuid.uuid4())

    def put(self, obj):
        """Put an object into the store."""
        id = self._genId()
        self._store[id] = obj
        return id

    def fetch(self, id):
        """Fetch an object from the store by id."""
        if id not in self._store:
            return None
        obj = self._store[id]
        if not hasattr(obj, "__remote_method"):
            del self._store[id]
        if hasattr(obj, "__jailed_pairs__"):
            _id = getKeyByValue(self._store, obj.__jailed_pairs__)
            self.fetch(_id)
        return obj


def task_worker(self, q, logger, abort):
    """Implement a task worker."""
    while True:
        try:
            if abort is not None and abort.is_set():
                break
            d = q.get()
            q.task_done()
            if d is None:
                continue
            if d["type"] == "getInterface":
                self._sendInterface()
            elif d["type"] == "setInterface":
                self._setRemote(d["api"])
                self.emit({"type": "interfaceSetAsRemote"})
                if not self._init:
                    self.emit({"type": "getInterface"})
                    self._init = True
            elif d["type"] == "interfaceSetAsRemote":
                # self.emit({'type':'getInterface'})
                self._remote_set = True
            elif d["type"] == "execute":
                if not self._executed:
                    try:
                        t = d["code"]["type"]
                        if t == "script":
                            content = d["code"]["content"]
                            exec(content, self._local)
                            self._executed = True
                        elif t == "requirements":
                            pass
                        else:
                            raise Exception("unsupported type")
                        self.emit({"type": "executeSuccess"})
                    except Exception as e:
                        logger.info(
                            "error during execution: %s", traceback.format_exc()
                        )
                        self.emit({"type": "executeFailure", "error": repr(e)})
            elif d["type"] == "method":
                interface = self._interface
                if "pid" in d and d["pid"] is not None:
                    interface = self._plugin_interfaces[d["pid"]]
                if d["name"] in interface:
                    if "promise" in d:
                        try:
                            resolve, reject = self._unwrap(d["promise"], False)
                            method = interface[d["name"]]
                            args = self._unwrap(d["args"], True)
                            # args.append({'id': self.id})
                            result = method(*args)
                            resolve(result)
                        except Exception as e:
                            logger.error(
                                "error in method %s: %s",
                                d["name"],
                                traceback.format_exc(),
                            )
                            reject(e)
                    else:
                        try:
                            method = interface[d["name"]]
                            args = self._unwrap(d["args"], True)
                            # args.append({'id': self.id})
                            method(*args)
                        except Exception:
                            logger.error(
                                "error in method %s: %s",
                                d["name"],
                                traceback.format_exc(),
                            )
                else:
                    raise Exception("method " + d["name"] + " is not found.")
            elif d["type"] == "callback":
                if "promise" in d:
                    resolve, reject = self._unwrap(d["promise"], False)
                    try:
                        method = self._store.fetch(d["num"])
                        if method is None:
                            raise Exception(
                                "Callback function can only called once, "
                                "if you want to call a function for multiple times, "
                                "please make it as a plugin api function. "
                                "See https://imjoy.io/docs for more details."
                            )
                        args = self._unwrap(d["args"], True)
                        result = method(*args)
                        resolve(result)
                    except Exception as e:
                        logger.error(
                            "error in method %s: %s", d["num"], traceback.format_exc()
                        )
                        reject(e)
                else:
                    try:
                        method = self._store.fetch(d["num"])
                        if method is None:
                            raise Exception(
                                "Callback function can only called once, "
                                "if you want to call a function for multiple times, "
                                "please make it as a plugin api function. "
                                "See https://imjoy.io/docs for more details."
                            )
                        args = self._unwrap(d["args"], True)
                        method(*args)
                    except Exception:
                        logger.error(
                            "error in method %s: %s", d["num"], traceback.format_exc()
                        )
            sys.stdout.flush()
        except queue.Empty:
            time.sleep(0.1)


class Promise(object):
    """Represent a promise."""

    def __init__(self, pfunc):
        """Set up promise."""
        self._resolve_handler = None
        self._finally_handler = None
        self._catch_handler = None

        def resolve(*args, **kwargs):
            self.resolve(*args, **kwargs)

        def reject(*args, **kwargs):
            self.reject(*args, **kwargs)

        pfunc(resolve, reject)

    def resolve(self, result):
        """Resolve promise."""
        try:
            if self._resolve_handler:
                self._resolve_handler(result)
        except Exception as e:
            if self._catch_handler:
                self._catch_handler(e)
            elif not self._finally_handler:
                print("Uncaught Exception: " + str(e))
        finally:
            if self._finally_handler:
                self._finally_handler()

    def reject(self, error):
        """Reject promise."""
        try:
            if self._catch_handler:
                self._catch_handler(error)
            elif not self._finally_handler:
                print("Uncaught Exception: " + str(error))
        finally:
            if self._finally_handler:
                self._finally_handler()

    def then(self, handler):
        """Implement then callback.

        Set handler and return the promise.
        """
        self._resolve_handler = handler
        return self

    def finally_(self, handler):
        """Implement finally callback.

        Set handler and return the promise.
        """
        self._finally_handler = handler
        return self

    def catch(self, handler):
        """Implement catch callback.

        Set handler and return the promise.
        """
        self._catch_handler = handler
        return self



import logging
import time
from unittest import TestCase
from copy import deepcopy

from .. import SocketIO, LoggingNamespace, find_callback


HOST = "localhost"
PORT = 9000
DATA = "xxx"
PAYLOAD = {"xxx": "yyy"}
BIN_DATA = bytearray(b"\xff\xff\xff")
BIN_PAYLOAD = {"data": BIN_DATA, "array": [bytearray(b"\xee"), bytearray(b"\xdd")]}
logging.basicConfig(level=logging.DEBUG)


class BaseMixin(object):
    def setUp(self):
        super(BaseMixin, self).setUp()
        self.called_on_response = False
        self.wait_time_in_seconds = 1

    def tearDown(self):
        super(BaseMixin, self).tearDown()
        self.socketIO.disconnect()

    def test_disconnect(self):
        "Disconnect"
        namespace = self.socketIO.define(Namespace)
        self.assertTrue(self.socketIO.connected)
        self.assertFalse(namespace.called_on_disconnect)
        self.socketIO.disconnect()
        self.assertTrue(namespace.called_on_disconnect)
        self.assertFalse(self.socketIO.connected)

    def test_emit(self):
        "Emit"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.emit("emit")
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(namespace.args_by_event, {"emit_response": ()})

    def test_emit_with_payload(self):
        "Emit with payload"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.emit("emit_with_payload", deepcopy(PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            namespace.args_by_event, {"emit_with_payload_response": (PAYLOAD,)}
        )

    def test_emit_with_multiple_payloads(self):
        "Emit with multiple payloads"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.emit(
            "emit_with_multiple_payloads", deepcopy(PAYLOAD), deepcopy(PAYLOAD)
        )
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            namespace.args_by_event,
            {"emit_with_multiple_payloads_response": (PAYLOAD, PAYLOAD)},
        )

    def test_emit_with_binary_payload(self):
        "Emit with binary payload"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.emit("emit_with_payload", deepcopy(BIN_PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            namespace.args_by_event, {"emit_with_payload_response": (BIN_PAYLOAD,)}
        )

    def test_emit_with_callback(self):
        "Emit with callback"
        self.socketIO.emit("emit_with_callback", self.on_response)
        self.socketIO.wait_for_callbacks(seconds=self.wait_time_in_seconds)
        self.assertTrue(self.called_on_response)

    def test_emit_with_callback_with_payload(self):
        "Emit with callback with payload"
        self.socketIO.emit("emit_with_callback_with_payload", self.on_response)
        self.socketIO.wait_for_callbacks(seconds=self.wait_time_in_seconds)
        self.assertTrue(self.called_on_response)

    def test_emit_with_callback_with_multiple_payloads(self):
        "Emit with callback with multiple payloads"
        self.socketIO.emit(
            "emit_with_callback_with_multiple_payloads", self.on_response
        )
        self.socketIO.wait_for_callbacks(seconds=self.wait_time_in_seconds)
        self.assertTrue(self.called_on_response)

    def test_emit_with_callback_with_binary_payload(self):
        "Emit with callback with binary payload"
        self.socketIO.emit(
            "emit_with_callback_with_binary_payload", self.on_binary_response
        )
        self.socketIO.wait_for_callbacks(seconds=self.wait_time_in_seconds)
        self.assertTrue(self.called_on_response)

    def test_emit_with_event(self):
        "Emit to trigger an event"
        self.socketIO.on("emit_with_event_response", self.on_response)
        self.socketIO.emit("emit_with_event", PAYLOAD)
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertTrue(self.called_on_response)

    def test_send(self):
        "Send"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.send()
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(namespace.response, "message_response")

    def test_send_with_data(self):
        "Send with data"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.send(deepcopy(DATA))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(namespace.response, DATA)

    def test_send_with_binary_data(self):
        "Send with binary data"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.send(deepcopy(BIN_DATA))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(namespace.response, BIN_DATA)

    def test_ack(self):
        "Respond to a server callback request"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.emit("trigger_server_expects_callback", deepcopy(PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            namespace.args_by_event,
            {
                "server_expects_callback": (PAYLOAD,),
                "server_received_callback": (PAYLOAD,),
            },
        )

    def test_binary_ack(self):
        "Respond to a server callback request with binary data"
        namespace = self.socketIO.define(Namespace)
        self.socketIO.emit("trigger_server_expects_callback", deepcopy(BIN_PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            namespace.args_by_event,
            {
                "server_expects_callback": (BIN_PAYLOAD,),
                "server_received_callback": (BIN_PAYLOAD,),
            },
        )

    def test_wait_with_disconnect(self):
        "Exit loop when the client wants to disconnect"
        self.socketIO.define(Namespace)
        self.socketIO.disconnect()
        timeout_in_seconds = 5
        start_time = time.time()
        self.socketIO.wait(timeout_in_seconds)
        self.assertTrue(time.time() - start_time < timeout_in_seconds)

    def test_namespace_emit(self):
        "Behave differently in different namespaces"
        main_namespace = self.socketIO.define(Namespace)
        chat_namespace = self.socketIO.define(Namespace, "/chat")
        news_namespace = self.socketIO.define(Namespace, "/news")
        news_namespace.emit("emit_with_payload", deepcopy(PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(main_namespace.args_by_event, {})
        self.assertEqual(chat_namespace.args_by_event, {})
        self.assertEqual(
            news_namespace.args_by_event, {"emit_with_payload_response": (PAYLOAD,)}
        )

    def test_namespace_emit_with_binary(self):
        "Make sure packet encoding works correctly"
        main_namespace = self.socketIO.define(Namespace)
        chat_namespace = self.socketIO.define(Namespace, "/chat")
        news_namespace = self.socketIO.define(Namespace, "/news")
        news_namespace.emit("emit_with_payload", deepcopy(BIN_PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(main_namespace.args_by_event, {})
        self.assertEqual(chat_namespace.args_by_event, {})
        self.assertEqual(
            news_namespace.args_by_event, {"emit_with_payload_response": (BIN_PAYLOAD,)}
        )

    def test_namespace_ack(self):
        "Respond to a server callback request within a namespace"
        chat_namespace = self.socketIO.define(Namespace, "/chat")
        chat_namespace.emit("trigger_server_expects_callback", deepcopy(PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            chat_namespace.args_by_event,
            {
                "server_expects_callback": (PAYLOAD,),
                "server_received_callback": (PAYLOAD,),
            },
        )

    def test_namespace_ack_with_binary(self):
        "Respond to a server callback request within a namespace"
        chat_namespace = self.socketIO.define(Namespace, "/chat")
        chat_namespace.emit("trigger_server_expects_callback", deepcopy(BIN_PAYLOAD))
        self.socketIO.wait(self.wait_time_in_seconds)
        self.assertEqual(
            chat_namespace.args_by_event,
            {
                "server_expects_callback": (BIN_PAYLOAD,),
                "server_received_callback": (BIN_PAYLOAD,),
            },
        )

    def on_response(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                self.assertEqual(arg, PAYLOAD)
            else:
                self.assertEqual(arg, DATA)
        self.called_on_response = True

    def on_binary_response(self, *args):
        for arg in args:
            if isinstance(arg, dict):
                self.assertEqual(arg, BIN_PAYLOAD)
            else:
                self.assertEqual(arg, BIN_DATA)
        self.called_on_response = True


class Test_XHR_PollingTransport(BaseMixin, TestCase):
    def setUp(self):
        super(Test_XHR_PollingTransport, self).setUp()
        self.socketIO = SocketIO(
            HOST, PORT, LoggingNamespace, transports=["xhr-polling"], verify=False
        )
        self.assertEqual(self.socketIO.transport_name, "xhr-polling")


class Test_WebsocketTransport(BaseMixin, TestCase):
    def setUp(self):
        super(Test_WebsocketTransport, self).setUp()
        self.socketIO = SocketIO(
            HOST,
            PORT,
            LoggingNamespace,
            transports=["xhr-polling", "websocket"],
            verify=False,
        )
        self.assertEqual(self.socketIO.transport_name, "websocket")


class Namespace(LoggingNamespace):
    def initialize(self):
        self.called_on_disconnect = False
        self.args_by_event = {}
        self.response = None

    def on_disconnect(self):
        self.called_on_disconnect = True

    def on_wait_with_disconnect_response(self):
        self.disconnect()

    def on_event(self, event, *args):
        callback, args = find_callback(args)
        if callback:
            callback(*args)
        self.args_by_event[event] = args

    def on_message(self, data):
        self.response = data



import six
import json
from collections import namedtuple
from base64 import b64encode
from copy import deepcopy

from .symmetries import (
    decode_string,
    encode_string,
    get_byte,
    get_character,
    get_int,
    parse_url,
)


EngineIOSession = namedtuple(
    "EngineIOSession", ["id", "ping_interval", "ping_timeout", "transport_upgrades"]
)


class SocketIOPacket:
    def __init__(self, packet_type, path, ack_id, args, attachments):
        self.type = packet_type
        self.path = path
        self.ack_id = ack_id
        self.args = args
        self.attachments = attachments
        self.binary_packets = []

    def __repr__(self):
        return (
            "<SocketIOPacket type: %s, path: %s, ack_id: %s, args: %s, attach: %s>"
            % (self.type, self.path, self.ack_id, self.args, self.attachments)
        )

    @property
    def finished(self):
        return self.attachments == len(self.binary_packets)

    def add(self, packet):
        self.binary_packets.append(packet)
        if self.finished:
            self.replace_placeholders()

    def replace_placeholders(self):
        def predicate(obj):
            if type(obj) is dict:
                return "_placeholder" in obj and "num" in obj
            else:
                return False

        def fn(obj):
            return bytearray(self.binary_packets[obj["num"]])

        self.args = traverse(deepcopy(self.args), predicate, fn)


def traverse(obj, predicate, fn):
    if predicate(obj):
        return fn(obj)
    elif isinstance(obj, dict):
        for key, value in six.iteritems(obj):
            obj[key] = traverse(value, predicate, fn)
        return obj
    elif isinstance(obj, (tuple, list)):
        obj = list(obj)
        for i, value in enumerate(obj):
            obj[i] = traverse(value, predicate, fn)
        return obj
    else:
        return obj


def parse_host(host, port, resource):
    if not host.startswith("http"):
        host = "http://" + host
    url_pack = parse_url(host)
    is_secure = url_pack.scheme == "https"
    port = port or url_pack.port or (443 if is_secure else 80)
    url = "%s:%d%s/%s" % (url_pack.hostname, port, url_pack.path, resource)
    return is_secure, url


def parse_engineIO_session(engineIO_packet_data):
    d = json.loads(decode_string(engineIO_packet_data))
    return EngineIOSession(
        id=d["sid"],
        ping_interval=d["pingInterval"] / float(1000),
        ping_timeout=d["pingTimeout"] / float(1000),
        transport_upgrades=d["upgrades"],
    )


def encode_engineIO_content(engineIO_packets):
    content = bytearray()
    for packet_type, packet_data in engineIO_packets:
        packet_text = format_packet_text(packet_type, packet_data)
        content.extend(_make_packet_prefix(packet_text) + packet_text)
    return content


def decode_engineIO_content(content):
    content_index = 0
    content_length = len(content)
    while content_index < content_length:
        content_index, packet_length = _read_packet_length(content, content_index)
        content_index, packet_text = _read_packet_text(
            content, content_index, packet_length
        )
        engineIO_packet_type, engineIO_packet_data = parse_packet_text(packet_text)
        yield engineIO_packet_type, engineIO_packet_data


def format_socketIO_packet_data(path=None, ack_id=None, args=None):
    binary_packets = []

    def predicate(obj):
        return isinstance(obj, bytearray)

    def fn(data):
        binary_packets.append(bytearray(b64encode(six.binary_type(data))))
        return {"_placeholder": True, "num": len(binary_packets) - 1}

    args = traverse(deepcopy(args), predicate, fn)
    socketIO_packet_data = json.dumps(args, ensure_ascii=False) if args else ""
    if ack_id is not None:
        socketIO_packet_data = str(ack_id) + socketIO_packet_data
    if path:
        socketIO_packet_data = path + "," + socketIO_packet_data
    if binary_packets:
        socketIO_packet_data = "%s-%s" % (len(binary_packets), socketIO_packet_data)
    return socketIO_packet_data, binary_packets


def parse_socketIO_packet(socketIO_packet):
    packet_type = get_int(socketIO_packet, 0)

    packet = decode_string(socketIO_packet[1:])
    # Binary types
    if packet_type in [5, 6]:
        attachments, packet = packet.split("-", 1)
    else:
        attachments = "0"

    if packet.startswith("/"):
        try:
            path, packet = packet.split(",", 1)
        except ValueError:
            path = packet
            packet = ""
    else:
        path = ""

    try:
        ack_id_string, packet = packet.split("[", 1)
        packet = "[" + packet
        ack_id = int(ack_id_string)
    except (ValueError, IndexError):
        ack_id = None

    try:
        args = json.loads(packet)
    except ValueError:
        args = []
    return SocketIOPacket(packet_type, path, ack_id, args, int(attachments))


def format_packet_text(packet_type, packet_data):
    if isinstance(packet_data, bytearray):
        packet_data = packet_data.decode("utf-8")
    return encode_string(str(packet_type) + packet_data)


def parse_packet_text(packet_text):
    packet_type = get_int(packet_text, 0)
    packet_data = packet_text[1:]
    return packet_type, packet_data


def get_namespace_path(socketIO_packet_data):
    if not socketIO_packet_data.startswith(b"/"):
        return ""
    # Loop incrementally in case there is binary data
    parts = []
    for i in range(len(socketIO_packet_data)):
        character = get_character(socketIO_packet_data, i)
        if "," == character:
            break
        parts.append(character)
    return "".join(parts)


def _make_packet_prefix(packet):
    length_string = str(len(packet))
    header_digits = bytearray([0])
    for i in range(len(length_string)):
        header_digits.append(ord(length_string[i]) - 48)
    header_digits.append(255)
    return header_digits


def _read_packet_length(content, content_index):
    while get_byte(content, content_index) not in [0, 1]:
        content_index += 1
    content_index += 1
    packet_length_string = ""
    byte = get_byte(content, content_index)
    while byte != 255:
        packet_length_string += str(byte)
        content_index += 1
        byte = get_byte(content, content_index)
    return content_index, int(packet_length_string)


def _read_packet_text(content, content_index, packet_length):
    while get_byte(content, content_index) == 255:
        content_index += 1
    packet_text = content[content_index : content_index + packet_length]
    return content_index + packet_length, packet_text



import logging
from threading import Thread, Event

from .exceptions import ConnectionError, TimeoutError


class HeartbeatThread(Thread):

    daemon = True

    def __init__(
        self, send_heartbeat, relax_interval_in_seconds, hurry_interval_in_seconds
    ):
        super(HeartbeatThread, self).__init__()
        self._send_heartbeat = send_heartbeat
        self._relax_interval_in_seconds = relax_interval_in_seconds
        self._hurry_interval_in_seconds = hurry_interval_in_seconds
        self._adrenaline = Event()
        self._rest = Event()
        self._halt = Event()

    def run(self):
        try:
            while not self._halt.is_set():
                try:
                    self._send_heartbeat()
                except TimeoutError:
                    pass
                if self._adrenaline.is_set():
                    interval_in_seconds = self._hurry_interval_in_seconds
                else:
                    interval_in_seconds = self._relax_interval_in_seconds
                self._rest.wait(interval_in_seconds)
        except ConnectionError:
            logging.debug("[heartbeat connection error]")

    def relax(self):
        self._adrenaline.clear()

    def hurry(self):
        self._adrenaline.set()
        self._rest.set()
        self._rest.clear()

    @property
    def hurried(self):
        return self._adrenaline.is_set()

    def halt(self):
        self._rest.set()
        self._halt.set()



"""Provide utils for Python 3 plugins."""
import asyncio
import sys
import traceback
import inspect

from imjoyUtils import Promise


async def task_worker(self, async_q, logger, abort=None):
    """Implement a task worker."""
    while True:
        if abort is not None and abort.is_set():
            break
        d = await async_q.get()
        try:
            if d is None:
                continue
            if d["type"] == "getInterface":
                self._sendInterface()
            elif d["type"] == "setInterface":
                self._setRemote(d["api"])
                self.emit({"type": "interfaceSetAsRemote"})
                if not self._init:
                    self.emit({"type": "getInterface"})
                    self._init = True
            elif d["type"] == "interfaceSetAsRemote":
                # self.emit({'type':'getInterface'})
                self._remote_set = True
            elif d["type"] == "execute":
                if not self._executed:
                    try:
                        t = d["code"]["type"]
                        if t == "script":
                            content = d["code"]["content"]
                            exec(content, self._local)
                            self._executed = True
                        elif t == "requirements":
                            pass
                        else:
                            raise Exception("unsupported type")
                        self.emit({"type": "executeSuccess"})
                    except Exception as e:
                        logger.info(
                            "error during execution: %s", traceback.format_exc()
                        )
                        self.emit({"type": "executeFailure", "error": repr(e)})
            elif d["type"] == "method":
                if d["name"] in self._interface:
                    if "promise" in d:
                        try:
                            resolve, reject = self._unwrap(d["promise"], False)
                            method = self._interface[d["name"]]
                            args = self._unwrap(d["args"], True)
                            # args.append({'id': self.id})
                            result = method(*args)
                            if result is not None and inspect.isawaitable(result):
                                result = await result
                            resolve(result)
                        except Exception as e:
                            logger.error(
                                "error in method %s: %s",
                                d["name"],
                                traceback.format_exc(),
                            )
                            reject(e)
                    else:
                        try:
                            method = self._interface[d["name"]]
                            args = self._unwrap(d["args"], True)
                            # args.append({'id': self.id})
                            result = method(*args)
                            if result is not None and inspect.isawaitable(result):
                                await result
                        except Exception:
                            logger.error(
                                "error in method %s: %s",
                                d["name"],
                                traceback.format_exc(),
                            )
                else:
                    raise Exception("method " + d["name"] + " is not found.")
            elif d["type"] == "callback":
                if "promise" in d:
                    resolve, reject = self._unwrap(d["promise"], False)
                    try:
                        method = self._store.fetch(d["num"])
                        if method is None:
                            raise Exception(
                                "Callback function can only called once, "
                                "if you want to call a function for multiple times, "
                                "please make it as a plugin api function. "
                                "See https://imjoy.io/docs for more details."
                            )
                        args = self._unwrap(d["args"], True)
                        result = method(*args)
                        if result is not None and inspect.isawaitable(result):
                            result = await result
                        resolve(result)
                    except Exception as e:
                        logger.error(
                            "error in method %s: %s", d["num"], traceback.format_exc()
                        )
                        reject(e)
                else:
                    try:
                        method = self._store.fetch(d["num"])
                        if method is None:
                            raise Exception(
                                "Callback function can only called once, "
                                "if you want to call a function for multiple times, "
                                "please make it as a plugin api function. "
                                "See https://imjoy.io/docs for more details."
                            )
                        args = self._unwrap(d["args"], True)
                        result = method(*args)
                        if result is not None and inspect.isawaitable(result):
                            await result
                    except Exception:
                        logger.error(
                            "error in method %s: %s", d["num"], traceback.format_exc()
                        )
        except Exception:
            print("error occured in the loop.", traceback.format_exc())
        finally:
            sys.stdout.flush()
            async_q.task_done()


class FuturePromise(Promise, asyncio.Future):
    """Represent a promise as a future."""

    def __init__(self, pfunc, loop):
        """Set up promise."""
        self.loop = loop
        Promise.__init__(self, pfunc)
        asyncio.Future.__init__(self)

    def resolve(self, result):
        """Resolve promise."""
        if self._resolve_handler or self._finally_handler:
            Promise.resolve(self, result)
        else:
            self.loop.call_soon(self.set_result, result)

    def reject(self, error):
        """Reject promise."""
        if self._catch_handler or self._finally_handler:
            Promise.reject(self, error)
        else:
            if error:
                self.loop.call_soon(self.set_exception, Exception())
            else:
                self.loop.call_soon(self.set_exception, Exception(str(error)))



import logging
import time
from .symmetries import NullHandler

LOGGER = logging.getLogger("socketIO-client")
LOGGER.addHandler(NullHandler())


class LoggingMixin(object):
    def _log(self, level, msg, *attrs):
        LOGGER.log(level, "%s %s" % (self._log_name, msg), *attrs)

    def _debug(self, msg, *attrs):
        self._log(logging.DEBUG, msg, *attrs)

    def _info(self, msg, *attrs):
        self._log(logging.INFO, msg, *attrs)

    def _warn(self, msg, *attrs):
        self._log(logging.WARNING, msg, *attrs)

    def _yield_warning_screen(self, seconds=None):
        last_warning = None
        for elapsed_time in _yield_elapsed_time(seconds):
            try:
                yield elapsed_time
            except Exception as warning:
                warning = str(warning)
                if last_warning != warning:
                    last_warning = warning
                    self._warn(warning)
                time.sleep(1)


def _yield_elapsed_time(seconds=None):
    start_time = time.time()
    if seconds is None:
        while True:
            yield _get_elapsed_time(start_time)
    while _get_elapsed_time(start_time) < seconds:
        yield _get_elapsed_time(start_time)


def _get_elapsed_time(start_time):
    return time.time() - start_time



"""Implement the ImJoy plugin engine."""
import argparse
import asyncio
import datetime
import json
import logging
import os
import pathlib
import platform
import shlex
import shutil
import string
import subprocess
import sys
import threading
import time
import traceback
import uuid
from mimetypes import MimeTypes
from urllib.parse import urlparse

import aiohttp_cors
import socketio
import yaml

# import webbrowser
from aiohttp import streamer, web

if sys.platform == "win32":
    from ctypes import windll

    def get_drives():
        """Return windows drives."""
        drives = []
        bitmask = windll.kernel32.GetLogicalDrives()
        for letter in string.ascii_uppercase:
            if bitmask & 1:
                drives.append(os.path.abspath(letter + ":/"))
            bitmask >>= 1
        return drives


try:
    import psutil
except Exception:
    print(
        "WARNING: a library called 'psutil' can not be imported, "
        "this may cause problem when killing processes."
    )


# read version information from file
HERE = pathlib.Path(__file__).parent
version_info = json.loads((HERE / "VERSION").read_text())

__version__ = version_info["version"]
API_VERSION = version_info["api_version"]

CONDA_AVAILABLE = False
MAX_ATTEMPTS = 1000
NAME_SPACE = "/"

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("ImJoyPluginEngine")

# add executable path to PATH
os.environ["PATH"] = (
    os.path.split(sys.executable)[0] + os.pathsep + os.environ.get("PATH", "")
)


try:
    process = subprocess.Popen(
        ["conda", "info", "--json", "-s"], stdout=subprocess.PIPE
    )
    cout, err = process.communicate()
    conda_prefix = json.loads(cout.decode("ascii"))["conda_prefix"]
    logger.info("Found conda environment: %s", conda_prefix)
    # for fixing CondaHTTPError:
    # https://github.com/conda/conda/issues/6064#issuecomment-458389796
    if os.name == "nt":
        os.environ["PATH"] = (
            os.path.join(conda_prefix, "Library", "bin")
            + os.pathsep
            + os.environ["PATH"]
        )
    CONDA_AVAILABLE = True
except OSError:
    conda_prefix = None
    if sys.version_info < (3, 0):
        sys.exit(
            "Sorry, ImJoy plugin engine can only run within a conda environment "
            "or at least in Python 3."
        )
    print(
        "WARNING: you are running ImJoy without conda, "
        "you may have problem with some plugins."
    )

parser = argparse.ArgumentParser()
parser.add_argument("--token", type=str, default=None, help="connection token")
parser.add_argument("--debug", action="store_true", help="debug mode")
parser.add_argument(
    "--serve", action="store_true", help="download ImJoy web app and serve it locally"
)
parser.add_argument("--host", type=str, default="127.0.0.1", help="socketio host")
parser.add_argument(
    "--base_url",
    type=str,
    default=None,
    help="the base url for accessing this plugin engine",
)
parser.add_argument("--port", type=str, default="9527", help="socketio port")
parser.add_argument(
    "--force_quit_timeout",
    type=int,
    default=5,
    help="the time (in second) for waiting before kill a plugin process, default: 5 s",
)
parser.add_argument(
    "--workspace",
    type=str,
    default="~/ImJoyWorkspace",
    help="workspace folder for plugins",
)
parser.add_argument(
    "--freeze", action="store_true", help="disable conda and pip commands"
)
parser.add_argument(
    "--engine_container_token",
    type=str,
    default=None,
    help="A token set by the engine container which launches the engine",
)

opt = parser.parse_args()

if opt.base_url is None or opt.base_url == "":
    opt.base_url = "http://{}:{}".format(opt.host, opt.port)

if opt.base_url.endswith("/"):
    opt.base_url = opt.base_url[:-1]

if not CONDA_AVAILABLE and not opt.freeze:
    print(
        "WARNING: `pip install` command may not work, "
        "in that case you may want to add `--freeze`."
    )

if opt.freeze:
    print(
        "WARNING: you are running the plugin engine with `--freeze`, "
        "this means you need to handle all the plugin requirements yourself."
    )

FORCE_QUIT_TIMEOUT = opt.force_quit_timeout
WORKSPACE_DIR = os.path.expanduser(opt.workspace)
if not os.path.exists(WORKSPACE_DIR):
    os.makedirs(WORKSPACE_DIR)

# read token from file if exists
try:
    if opt.token is None or opt.token == "":
        with open(os.path.join(WORKSPACE_DIR, ".token"), "r") as f:
            opt.token = f.read()
except Exception:
    logger.debug("Failed to read token from file")

try:
    if opt.token is None or opt.token == "":
        opt.token = str(uuid.uuid4())
        with open(os.path.join(WORKSPACE_DIR, ".token"), "w") as f:
            f.write(opt.token)
except Exception as e:
    logger.error("Failed to save .token file: %s", str(e))


def killProcess(pid):
    """Kill process."""
    try:
        cp = psutil.Process(pid)
        for proc in cp.children(recursive=True):
            try:
                if proc.is_running():
                    proc.kill()
            except psutil.NoSuchProcess:
                logger.info("subprocess %s has already been killed", pid)
            except Exception as e:
                logger.error(
                    "WARNING: failed to kill a subprocess (PID={}). Error: {}".format(
                        pid, str(e)
                    )
                )
        cp.kill()
        logger.info("plugin process %s was killed.", pid)
    except psutil.NoSuchProcess:
        logger.info("process %s has already been killed", pid)
    except Exception as e:
        logger.error(
            "WARNING: failed to kill a process (PID={}), "
            "you may want to kill it manually. Error: {}".format(pid, str(e))
        )


# try to kill last process
pid_file = os.path.join(WORKSPACE_DIR, ".pid")
try:
    if os.path.exists(pid_file):
        with open(pid_file, "r") as f:
            killProcess(int(f.read()))
except Exception:
    logger.debug("Failed to kill last process")
try:
    pid = str(os.getpid())
    with open(pid_file, "w") as f:
        f.write(pid)
except Exception as e:
    logger.error("Failed to save .pid file: %s", str(e))

WEB_APP_DIR = os.path.join(WORKSPACE_DIR, "__ImJoy__")
if opt.serve:
    if shutil.which("git") is None:
        print("Installing git...")
        ret = subprocess.Popen(
            "conda install -y git && git clone -b gh-pages --depth 1 "
            "https://github.com/oeway/ImJoy".split(),
            shell=False,
        ).wait()
        if ret != 0:
            print(
                "Failed to install git, please check whether you have internet access."
            )
            sys.exit(3)
    if os.path.exists(WEB_APP_DIR) and os.path.isdir(WEB_APP_DIR):
        ret = subprocess.Popen(["git", "stash"], cwd=WEB_APP_DIR, shell=False).wait()
        if ret != 0:
            print("Failed to clean files locally.")
        ret = subprocess.Popen(
            ["git", "pull", "--all"], cwd=WEB_APP_DIR, shell=False
        ).wait()
        if ret != 0:
            print("Failed to pull files for serving offline.")
        ret = subprocess.Popen(
            ["git", "checkout", "gh-pages"], cwd=WEB_APP_DIR, shell=False
        ).wait()
        if ret != 0:
            print("Failed to checkout files from gh-pages.")
    if not os.path.exists(WEB_APP_DIR):
        print("Downloading files for serving ImJoy locally...")
        ret = subprocess.Popen(
            "git clone -b gh-pages --depth 1 "
            "https://github.com/oeway/ImJoy __ImJoy__".split(),
            shell=False,
            cwd=WORKSPACE_DIR,
        ).wait()
        if ret != 0:
            print(
                "Failed to download files, "
                "please check whether you have internet access."
            )
            sys.exit(4)

# ALLOWED_ORIGINS = [opt.base_url, 'http://imjoy.io', 'https://imjoy.io']
sio = socketio.AsyncServer()
app = web.Application()
sio.attach(app)

if opt.debug:
    logger.setLevel(logging.DEBUG)
else:
    logger.setLevel(logging.ERROR)


if opt.serve and os.path.exists(os.path.join(WEB_APP_DIR, "index.html")):

    async def index(request):
        """Serve the client-side application."""
        with open(os.path.join(WEB_APP_DIR, "index.html"), "r", encoding="utf-8") as f:
            return web.Response(text=f.read(), content_type="text/html")

    app.router.add_static("/static", path=str(os.path.join(WEB_APP_DIR, "static")))
    # app.router.add_static('/docs/', path=str(os.path.join(WEB_APP_DIR, 'docs')))

    async def docs_handler(request):
        """Handle docs."""
        raise web.HTTPFound(location="https://imjoy.io/docs")

    app.router.add_get("/docs", docs_handler, name="docs")
    print("A local version of Imjoy web app is available at " + opt.base_url)
else:

    async def index(request):
        """Return index."""
        return web.Response(
            body=(
                '<H1><a href="https://imjoy.io">ImJoy.IO</a></H1><p>'
                'You can run "python -m imjoy --serve" '
                "to serve ImJoy web app locally.</p>"
            ),
            content_type="text/html",
        )


app.router.add_get("/", index)


async def about(request):
    """Return about text."""
    params = request.rel_url.query
    if "token" in params:
        body = (
            "<H1>ImJoy Plugin Engine connection token: </H1><H3>"
            + params["token"]
            + "</H3><br>"
        )
        body += (
            "<p>You have to specify this token when you connect the ImJoy web app "
            "to this Plugin Engine. The token will be saved and automatically reused "
            "when you launch the App again. </p>"
        )
        body += "<br>"
        body += (
            "<p>Alternatively, you can launch a new ImJoy instance "
            "with the link below: </p>"
        )

        if opt.serve:
            body += (
                '<p><a href="'
                + opt.base_url
                + "/#/app?token="
                + params["token"]
                + '">Open ImJoy App</a></p>'
            )
        else:
            body += (
                '<p><a href="https://imjoy.io/#/app?token='
                + params["token"]
                + '">Open ImJoy App</a></p>'
            )

    else:
        if opt.serve:
            body = '<H1><a href="' + opt.base_url + '/#/app">Open ImJoy App</a></H1>'
        else:
            body = '<H1><a href="https://imjoy.io/#/app">Open ImJoy App</a></H1>'
    body += (
        "<H2>Please use the latest Google Chrome browser to run the ImJoy App."
        '</H2><a href="https://www.google.com/chrome/">Download Chrome</a><p>'
        "Note: Safari is not supported "
        "due to its restrictions on connecting to localhost. "
        "Currently, only FireFox and Chrome (preferred) are supported.</p>"
    )
    return web.Response(body=body, content_type="text/html")


app.router.add_get("/about", about)

attempt_count = 0
cmd_history = []
plugins = {}
plugin_sessions = {}
plugin_sids = {}
plugin_signatures = {}
clients = {}
client_sessions = {}
registered_sessions = {}
generatedUrls = {}
generatedUrlFiles = {}
requestUploadFiles = {}
requestUrls = {}

default_requirements_py2 = ["requests", "six", "websocket-client", "numpy", "psutil"]
default_requirements_py3 = [
    "requests",
    "six",
    "websocket-client",
    "janus",
    "numpy",
    "psutil",
]

script_dir = os.path.dirname(os.path.normpath(__file__))
template_script = os.path.abspath(os.path.join(script_dir, "imjoyWorkerTemplate.py"))

if CONDA_AVAILABLE:
    if sys.platform == "linux" or sys.platform == "linux2":
        # linux
        conda_activate = "/bin/bash -c 'source " + conda_prefix + "/bin/activate {}'"
    elif sys.platform == "darwin":
        # OS X
        conda_activate = "source activate {}"
    elif sys.platform == "win32":
        # Windows...
        conda_activate = "activate {}"
    else:
        conda_activate = "conda activate {}"
else:
    conda_activate = "{}"


def resumePluginSession(pid, session_id, plugin_signature):
    """Resume plugin session."""
    if pid in plugins:
        if session_id in plugin_sessions:
            plugin_sessions[session_id].append(plugins[pid])
        else:
            plugin_sessions[session_id] = [plugins[pid]]

    if plugin_signature in plugin_signatures:
        plugin_info = plugin_signatures[plugin_signature]
        logger.info("resuming plugin %s", pid)
        return plugin_info
    else:
        return None


def addClientSession(session_id, client_id, sid, base_url, workspace):
    """Add client session."""
    if client_id in clients:
        clients[client_id].append(sid)
        client_connected = True
    else:
        clients[client_id] = [sid]
        client_connected = False
    logger.info("adding client session %s", sid)
    registered_sessions[sid] = {
        "client": client_id,
        "session": session_id,
        "base_url": base_url,
        "workspace": workspace,
    }
    return client_connected


def disconnectClientSession(sid):
    """Disconnect client session."""
    if sid in registered_sessions:
        logger.info("disconnecting client session %s", sid)
        obj = registered_sessions[sid]
        client_id, session_id = obj["client"], obj["session"]
        del registered_sessions[sid]
        if client_id in clients and sid in clients[client_id]:
            clients[client_id].remove(sid)
            if len(clients[client_id]) == 0:
                del clients[client_id]
        if session_id in plugin_sessions:
            for plugin in plugin_sessions[session_id]:
                if "allow-detach" not in plugin["flags"]:
                    killPlugin(plugin["id"])
            del plugin_sessions[session_id]


def addPlugin(plugin_info, sid=None):
    """Add plugin."""
    pid = plugin_info["id"]
    session_id = plugin_info["session_id"]
    plugin_signatures[plugin_info["signature"]] = plugin_info
    plugins[pid] = plugin_info
    if session_id in plugin_sessions:
        plugin_sessions[session_id].append(plugin_info)
    else:
        plugin_sessions[session_id] = [plugin_info]

    if pid in plugins and sid is not None:
        plugin_sids[sid] = plugin_info
        plugin_info["sid"] = sid


def disconnectPlugin(sid):
    """Disconnect plugin."""
    if sid in plugin_sids:
        logger.info("disconnecting plugin session %s", sid)
        pid = plugin_sids[sid]["id"]
        if pid in plugins:
            logger.info("clean up plugin %s", pid)
            if plugins[pid]["signature"] in plugin_signatures:
                logger.info("clean up plugin signature %s", plugins[pid]["signature"])
                del plugin_signatures[plugins[pid]["signature"]]
            del plugins[pid]
        del plugin_sids[sid]
        for session_id in plugin_sessions.keys():
            exist = False
            for p in plugin_sessions[session_id]:
                if p["id"] == pid:
                    exist = p
            if exist:
                logger.info("clean up plugin session %s", session_id)
                plugin_sessions[session_id].remove(exist)
                killPlugin(exist["id"])


def setPluginPID(plugin_id, pid):
    """Set plugin pid."""
    plugins[plugin_id]["process_id"] = pid


def killPlugin(pid):
    """Kill plugin."""
    if pid in plugins:
        try:
            plugins[pid]["abort"].set()
            plugins[pid]["aborting"] = asyncio.get_event_loop().create_future()
            killProcess(plugins[pid]["process_id"])
            print('INFO: "{}" was killed.'.format(pid))
        except Exception as e:
            print('WARNING: failed to kill plugin "{}".'.format(pid))
            logger.error(str(e))
        if "sid" in plugins[pid]:
            if plugins[pid]["sid"] in plugin_sids:
                del plugin_sids[plugins[pid]["sid"]]

        if plugins[pid]["signature"] in plugin_signatures:
            logger.info(
                "clean up killed plugin signature %s", plugins[pid]["signature"]
            )
            del plugin_signatures[plugins[pid]["signature"]]
        logger.info("clean up killed plugin %s", pid)
        del plugins[pid]


async def killAllPlugins(ssid):
    """Kill all plugins."""
    tasks = []
    for sid in list(plugin_sids.keys()):
        try:
            await on_kill_plugin(ssid, {"id": plugin_sids[sid]["id"]})
        except Exception as e:
            logger.error(str(e))

    return asyncio.gather(*tasks)


def parseRepos(requirements, work_dir):
    """Return a list of repositories from a list of requirements."""
    repos = []
    if type(requirements) is list:
        requirements = [str(r) for r in requirements]
        for r in requirements:
            if ":" in r:
                rs = r.split(":")
                tp, libs = rs[0], ":".join(rs[1:])
                tp, libs = tp.strip(), libs.strip()
                libs = [l.strip() for l in libs.split(" ") if l.strip() != ""]
                if tp == "repo" and len(libs) > 0:
                    name = libs[0].split("/")[-1].replace(".git", "")
                    repo = {
                        "url": libs[0],
                        "repo_dir": os.path.join(
                            work_dir, libs[1] if len(libs) > 1 else name
                        ),
                    }
                    repos.append(repo)
    return repos


def parseRequirements(requirements, default_requirements, work_dir):
    """Parse requirements."""
    requirements_cmd = "pip install " + " ".join(default_requirements)
    if type(requirements) is list:
        requirements = [str(r) for r in requirements]

        for r in requirements:
            if ":" in r:
                rs = r.split(":")
                tp, libs = rs[0], ":".join(rs[1:])
                tp, libs = tp.strip(), libs.strip()
                libs = [l.strip() for l in libs.split(" ") if l.strip() != ""]
                if tp == "conda" and len(libs) > 0:
                    requirements_cmd += " && conda install -y " + " ".join(libs)
                elif tp == "pip" and len(libs) > 0:
                    requirements_cmd += " && pip install " + " ".join(libs)
                elif tp == "repo" and len(libs) > 0:
                    pass
                elif tp == "cmd" and len(libs) > 0:
                    requirements_cmd += " && " + " ".join(libs)
                elif "+" in tp or "http" in tp:
                    requirements_cmd += " && pip install " + r
                else:
                    raise Exception("Unsupported requirement type: " + tp)
            else:
                requirements_cmd += " && pip install " + r

    elif type(requirements) is str and requirements.strip() != "":
        requirements_cmd += " && " + requirements
    else:
        raise Exception("Unsupported requirements type.")
    return requirements_cmd


def console_to_str(s):
    """From pypa/pip project, pip.backwardwardcompat. License MIT."""
    try:
        return s.decode(sys.__stdout__.encoding)
    except UnicodeDecodeError:
        return s.decode("utf_8")
    except AttributeError:  # for tests, #13
        return s


def runCmd(
    cmd,
    shell=False,
    cwd=None,
    log_in_real_time=True,
    check_returncode=True,
    callback=None,
    plugin_id=None,
):
    """Run command.

    From https://github.com/vcs-python/libvcs/.
    """
    proc = subprocess.Popen(
        cmd,
        shell=shell,
        stderr=subprocess.PIPE,
        stdout=subprocess.PIPE,
        creationflags=0,
        bufsize=1,
        cwd=cwd,
    )
    if plugin_id is not None:
        setPluginPID(plugin_id, proc.pid)

    all_output = []
    code = None
    line = None
    while code is None:
        code = proc.poll()
        if callback and callable(callback):
            line = console_to_str(proc.stderr.read(128))
            if line:
                callback(output=line, timestamp=datetime.datetime.now())
    if callback and callable(callback):
        callback(output="\r", timestamp=datetime.datetime.now())

    lines = filter(None, (line.strip() for line in proc.stdout.readlines()))
    all_output = console_to_str(b"\n".join(lines))
    if code:
        stderr_lines = filter(None, (line.strip() for line in proc.stderr.readlines()))
        all_output = console_to_str(b"".join(stderr_lines))
    output = "".join(all_output)
    if code != 0 and check_returncode:
        raise Exception("Command failed with code {}: {}".format(code, cmd))
    return output


def parseEnv(env, work_dir, default_env_name):
    """Parse environment."""
    env_name = ""
    is_py2 = False
    envs = None

    if type(env) is str:
        env = None if env.strip() == "" else env

    if env is not None:
        if not opt.freeze and CONDA_AVAILABLE:
            if type(env) is str:
                envs = [env]
            else:
                envs = env
            for i, env in enumerate(envs):
                if "conda create" in env:
                    # if not env.startswith('conda'):
                    #     raise Exception('env command must start with conda')
                    if "python=2" in env:
                        is_py2 = True
                    parms = shlex.split(env)
                    if "-n" in parms:
                        env_name = parms[parms.index("-n") + 1]
                    elif "--name" in parms:
                        env_name = parms[parms.index("--name") + 1]
                    else:
                        env_name = default_env_name
                        envs[i] = env.replace(
                            "conda create", "conda create -n " + env_name
                        )

                    if "-y" not in parms:
                        envs[i] = env.replace("conda create", "conda create -y")

                if "conda env create" in env:
                    parms = shlex.split(env)
                    if "-f" in parms:
                        try:
                            env_file = os.path.join(
                                work_dir, parms[parms.index("-f") + 1]
                            )
                            with open(env_file, "r") as stream:
                                env_config = yaml.load(stream)
                                assert "name" in env_config
                                env_name = env_config["name"]
                        except Exception as e:
                            raise Exception(
                                "Failed to read the env name "
                                "from the specified env file: " + str(e)
                            )

                    else:
                        raise Exception(
                            "You should provided a environment file "
                            "via the `conda env create -f`"
                        )

        else:
            print(
                "WARNING: blocked env command: \n{}\n"
                "You may want to run it yourself.".format(env)
            )
            logger.warning(
                "env command is blocked because conda is not avaialbe "
                "or in `--freeze` mode: %s",
                env,
            )

    if env_name.strip() == "":
        env_name = None

    return env_name, envs, is_py2


@sio.on("connect", namespace=NAME_SPACE)
def connect(sid, environ):
    """Connect client."""
    logger.info("connect %s", sid)


@sio.on("init_plugin", namespace=NAME_SPACE)
async def on_init_plugin(sid, kwargs):
    """Initialize plugin."""
    try:
        if sid in registered_sessions:
            obj = registered_sessions[sid]
            client_id, session_id = obj["client"], obj["session"]
        else:
            logger.debug("client %s is not registered.", sid)
            return {"success": False}
        pid = kwargs["id"]
        config = kwargs.get("config", {})
        env = config.get("env")
        cmd = config.get("cmd", "python")
        pname = config.get("name")
        flags = config.get("flags", [])
        tag = config.get("tag", "")
        requirements = config.get("requirements", []) or []
        workspace = config.get("workspace", "default")
        work_dir = os.path.join(WORKSPACE_DIR, workspace)
        if not os.path.exists(work_dir):
            os.makedirs(work_dir)
        plugin_env = os.environ.copy()
        plugin_env["WORK_DIR"] = work_dir
        logger.info(
            "initialize the plugin. name=%s, id=%s, cmd=%s, workspace=%s",
            pname,
            pid,
            cmd,
            workspace,
        )

        if "single-instance" in flags:
            plugin_signature = "{}/{}".format(pname, tag)
            resume = True
        elif "allow-detach" in flags:
            plugin_signature = "{}/{}/{}/{}".format(client_id, workspace, pname, tag)
            resume = True
        else:
            plugin_signature = None
            resume = False

        if resume:
            plugin_info = resumePluginSession(pid, session_id, plugin_signature)
            if plugin_info is not None:
                if "aborting" in plugin_info:
                    logger.info("Waiting for plugin %s to abort", plugin_info["id"])
                    await plugin_info["aborting"]
                else:
                    logger.debug("plugin already initialized: %s", pid)
                    return {
                        "success": True,
                        "resumed": True,
                        "initialized": True,
                        "secret": plugin_info["secret"],
                        "work_dir": os.path.abspath(work_dir),
                    }
            else:
                logger.info(
                    "failed to resume single instance plugin: %s, %s",
                    pid,
                    plugin_signature,
                )

        secretKey = str(uuid.uuid4())
        abort = threading.Event()
        plugin_info = {
            "secret": secretKey,
            "id": pid,
            "abort": abort,
            "flags": flags,
            "session_id": session_id,
            "name": config["name"],
            "type": config["type"],
            "client_id": client_id,
            "signature": plugin_signature,
        }
        logger.info("Add plugin: %s", str(plugin_info))
        addPlugin(plugin_info)

        @sio.on("from_plugin_" + secretKey, namespace=NAME_SPACE)
        async def message_from_plugin(sid, kwargs):
            # print('forwarding message_'+secretKey, kwargs)
            if kwargs["type"] in [
                "initialized",
                "importSuccess",
                "importFailure",
                "executeSuccess",
                "executeFailure",
            ]:
                await sio.emit("message_from_plugin_" + secretKey, kwargs)
                logger.debug("message from %s", pid)
                if kwargs["type"] == "initialized":
                    addPlugin(plugin_info, sid)
            else:
                await sio.emit(
                    "message_from_plugin_" + secretKey,
                    {"type": "message", "data": kwargs},
                )

        @sio.on("message_to_plugin_" + secretKey, namespace=NAME_SPACE)
        async def message_to_plugin(sid, kwargs):
            # print('forwarding message_to_plugin_'+secretKey, kwargs)
            if kwargs["type"] == "message":
                await sio.emit("to_plugin_" + secretKey, kwargs["data"])
            logger.debug("message to plugin %s", secretKey)

        eloop = asyncio.get_event_loop()

        def stop_callback(success, message):
            if "aborting" in plugin_info:
                plugin_info["aborting"].set_result(success)
            message = str(message or "")
            message = message[:100] + (message[100:] and "..")
            logger.info(
                "disconnecting from plugin (success:%s, message: %s)",
                str(success),
                message,
            )
            coro = sio.emit(
                "message_from_plugin_" + secretKey,
                {
                    "type": "disconnected",
                    "details": {"success": success, "message": message},
                },
            )
            asyncio.run_coroutine_threadsafe(coro, eloop).result()

        def logging_callback(msg, type="info"):
            if msg == "":
                return
            coro = sio.emit(
                "message_from_plugin_" + secretKey,
                {"type": "logging", "details": {"value": msg, "type": type}},
            )
            asyncio.run_coroutine_threadsafe(coro, eloop).result()

        args = '{} "{}" --id="{}" --server={} --secret="{}" --namespace={}'.format(
            cmd,
            template_script,
            pid,
            "http://127.0.0.1:" + opt.port,
            secretKey,
            NAME_SPACE,
        )
        taskThread = threading.Thread(
            target=launch_plugin,
            args=[
                stop_callback,
                logging_callback,
                pid,
                pname,
                tag,
                env,
                requirements,
                args,
                work_dir,
                abort,
                pid,
                plugin_env,
            ],
        )
        taskThread.daemon = True
        taskThread.start()
        return {
            "success": True,
            "initialized": False,
            "secret": secretKey,
            "work_dir": os.path.abspath(work_dir),
        }

    except Exception:
        logger.error(traceback.format_exc())
        return {"success": False, "reason": traceback.format_exc()}


async def force_kill_timeout(t, obj):
    """Force kill plugin after timeout."""
    pid = obj["pid"]
    for i in range(int(t * 10)):
        if obj["force_kill"]:
            await asyncio.sleep(0.1)
        else:
            return
    try:
        logger.warning("Timeout, force quitting %s", pid)
        killPlugin(pid)
    finally:
        return


@sio.on("reset_engine", namespace=NAME_SPACE)
async def on_reset_engine(sid, kwargs):
    """Reset engine."""
    logger.info("kill plugin: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered"}

    global attempt_count
    global attempt_count
    global cmd_history
    global plugins
    global plugin_sessions
    global plugin_sids
    global plugin_signatures
    # global clients
    # global client_sessions
    # global registered_sessions
    global generatedUrls
    global generatedUrlFiles
    global requestUploadFiles
    global requestUrls

    await killAllPlugins(sid)

    attempt_count = 0
    cmd_history = []
    plugins = {}
    plugin_sessions = {}
    plugin_sids = {}
    plugin_signatures = {}
    # clients = {}
    # client_sessions = {}
    # registered_sessions = {}
    generatedUrls = {}
    generatedUrlFiles = {}
    requestUploadFiles = {}
    requestUrls = {}

    return {"success": True}


@sio.on("kill_plugin", namespace=NAME_SPACE)
async def on_kill_plugin(sid, kwargs):
    """Kill plugin."""
    logger.info("kill plugin: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered"}

    pid = kwargs["id"]
    if pid in plugins:
        if "killing" not in plugins[pid]:
            obj = {"force_kill": True, "pid": pid}
            plugins[pid]["killing"] = True

            def exited(result):
                obj["force_kill"] = False
                logger.info("Plugin %s exited normally.", pid)
                # kill the plugin now
                killPlugin(pid)

            await sio.emit(
                "to_plugin_" + plugins[pid]["secret"],
                {"type": "disconnect"},
                callback=exited,
            )
            await force_kill_timeout(FORCE_QUIT_TIMEOUT, obj)
    return {"success": True}


@sio.on("register_client", namespace=NAME_SPACE)
async def on_register_client(sid, kwargs):
    """Register client."""
    global attempt_count
    client_id = kwargs.get("id", str(uuid.uuid4()))
    workspace = kwargs.get("workspace", "default")
    session_id = kwargs.get("session_id", str(uuid.uuid4()))
    base_url = kwargs.get("base_url", opt.base_url)
    if base_url.endswith("/"):
        base_url = base_url[:-1]

    token = kwargs.get("token")
    if token != opt.token:
        logger.debug("token mismatch: %s != %s", token, opt.token)
        print("======== Connection Token: " + opt.token + " ========")
        if opt.engine_container_token is not None:
            await sio.emit(
                "message_to_container_" + opt.engine_container_token,
                {
                    "type": "popup_token",
                    "client_id": client_id,
                    "session_id": session_id,
                },
            )
        # try:
        #     webbrowser.open(
        #         'http://'+opt.host+':'+opt.port+'/about?token='+opt.token,
        #         new=0, autoraise=True)
        # except Exception as e:
        #     print('Failed to open the browser.')
        attempt_count += 1
        if attempt_count >= MAX_ATTEMPTS:
            logger.info("Client exited because max attemps exceeded: %s", attempt_count)
            sys.exit(100)
        return {"success": False}
    else:
        attempt_count = 0
        if addClientSession(session_id, client_id, sid, base_url, workspace):
            confirmation = True
            message = (
                "Another ImJoy session is connected to this Plugin Engine({}), "
                "allow a new session to connect?".format(base_url)
            )
        else:
            confirmation = False
            message = None

        logger.info("register client: %s", kwargs)

        engine_info = {"api_version": API_VERSION, "version": __version__}
        engine_info["platform"] = {
            "uname": ", ".join(platform.uname()),
            "machine": platform.machine(),
            "system": platform.system(),
            "processor": platform.processor(),
            "node": platform.node(),
        }

        return {
            "success": True,
            "confirmation": confirmation,
            "message": message,
            "engine_info": engine_info,
        }


def scandir(path, type=None, recursive=False):
    """Scan a directory for a type of files return a list of files found."""
    file_list = []
    for f in os.scandir(path):
        if f.name.startswith("."):
            continue
        if type is None or type == "file":
            if os.path.isdir(f.path):
                if recursive:
                    file_list.append(
                        {
                            "name": f.name,
                            "type": "dir",
                            "children": scandir(f.path, type, recursive),
                        }
                    )
                else:
                    file_list.append({"name": f.name, "type": "dir"})
            else:
                file_list.append({"name": f.name, "type": "file"})
        elif type == "directory":
            if os.path.isdir(f.path):
                file_list.append({"name": f.name})
    return file_list


@sio.on("list_dir", namespace=NAME_SPACE)
async def on_list_dir(sid, kwargs):
    """List files in directory."""
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered."}

    try:
        workspace_dir = os.path.join(
            WORKSPACE_DIR, registered_sessions[sid]["workspace"]
        )

        path = kwargs.get("path", workspace_dir)

        if not os.path.isabs(path):
            path = os.path.join(workspace_dir, path)
        path = os.path.normpath(os.path.expanduser(path))

        type = kwargs.get("type")
        recursive = kwargs.get("recursive", False)
        files_list = {"success": True}
        files_list["path"] = path
        files_list["name"] = os.path.basename(os.path.abspath(path))
        files_list["type"] = "dir"
        files_list["children"] = scandir(files_list["path"], type, recursive)

        if sys.platform == "win32" and os.path.abspath(path) == os.path.abspath("/"):
            files_list["drives"] = get_drives()

        return files_list
    except Exception as e:
        logger.error("list dir error: %s", str(e))
        return {"success": False, "error": str(e)}


@sio.on("remove_files", namespace=NAME_SPACE)
async def on_remove_files(sid, kwargs):
    """Remove files."""
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered."}
    logger.info("removing files: %s", kwargs)
    workspace_dir = os.path.join(WORKSPACE_DIR, registered_sessions[sid]["workspace"])
    path = kwargs.get("path", workspace_dir)
    if not os.path.isabs(path):
        path = os.path.join(workspace_dir, path)
    path = os.path.normpath(os.path.expanduser(path))
    type = kwargs.get("type")
    recursive = kwargs.get("recursive", False)

    if os.path.exists(path) and not os.path.isdir(path) and type == "file":
        try:
            os.remove(path)
            return {"success": True}
        except Exception as e:
            logger.error("remove files error: %s", str(e))
            return {"success": False, "error": str(e)}
    elif os.path.exists(path) and os.path.isdir(path) and type == "dir":
        try:
            if recursive:
                dirname, filename = os.path.split(path)
                shutil.move(path, os.path.join(dirname, "." + filename))
                # shutil.rmtree(path)
            else:
                os.rmdir(path)
            return {"success": True}
        except Exception as e:
            logger.error("remove files error: %s", str(e))
            return {"success": False, "error": str(e)}
    else:
        logger.error("remove files error: %s", "File not exists or type mismatch.")
        return {"success": False, "error": "File not exists or type mismatch."}


@streamer
async def file_sender(writer, file_path=None):
    """Read a large file chunk by chunk and send it through HTTP.

    Do not read the chunks into memory.
    """
    with open(file_path, "rb") as f:
        chunk = f.read(2 ** 16)
        while chunk:
            await writer.write(chunk)
            chunk = f.read(2 ** 16)


@sio.on("request_upload_url", namespace=NAME_SPACE)
async def on_request_upload_url(sid, kwargs):
    """Request upload url."""
    logger.info("requesting file upload url: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered"}

    urlid = str(uuid.uuid4())
    fileInfo = {
        "id": urlid,
        "overwrite": kwargs.get("overwrite", False),
        "workspace": registered_sessions[sid]["workspace"],
    }
    if "path" in kwargs:
        fileInfo["path"] = kwargs["path"]

    if "dir" in kwargs:
        path = os.path.expanduser(kwargs["dir"])
        if not os.path.isabs(path):
            path = os.path.join(WORKSPACE_DIR, fileInfo["workspace"], path)
        fileInfo["dir"] = path

    if "path" in fileInfo:
        path = fileInfo["path"]
        if "dir" in fileInfo:
            path = os.path.join(fileInfo["dir"], path)
        else:
            path = os.path.join(WORKSPACE_DIR, fileInfo["workspace"], path)

        if os.path.exists(path) and not kwargs.get("overwrite", False):
            return {"success": False, "error": "file already exist."}

    base_url = kwargs.get("base_url", registered_sessions[sid]["base_url"])
    url = "{}/upload/{}".format(base_url, urlid)
    requestUrls[url] = fileInfo
    requestUploadFiles[urlid] = fileInfo
    return {"success": True, "id": urlid, "url": url}


async def upload_file(request):
    """Upload file."""
    urlid = request.match_info["urlid"]  # Could be a HUGE file
    if urlid not in requestUploadFiles:
        raise web.HTTPForbidden(text="Invalid URL")

    fileInfo = requestUploadFiles[urlid]
    try:
        reader = await request.multipart()
        field = None
        while True:
            part = await reader.next()
            print(part, part.filename)
            if part.filename is None:
                continue
            field = part
            break
        filename = field.filename
        # You cannot rely on Content-Length if transfer is chunked.
        size = 0
        if "path" in fileInfo:
            path = fileInfo["path"]
        else:
            path = filename

        if "dir" in fileInfo:
            path = os.path.join(fileInfo["dir"], path)
        else:
            path = os.path.join(WORKSPACE_DIR, fileInfo["workspace"], path)

        if os.path.exists(path) and not fileInfo.get("overwrite", False):
            return web.Response(body="File {} already exists.".format(path), status=404)

        logger.info("uploading file to %s", path)
        directory, _ = os.path.split(path)
        if not os.path.exists(directory):
            os.makedirs(directory)
        with open(path, "wb") as f:
            while True:
                chunk = await field.read_chunk()  # 8192 bytes by default.
                if not chunk:
                    break
                size += len(chunk)
                f.write(chunk)
        fileInfo["size"] = size
        fileInfo["path"] = path
        logger.info("file saved to %s (size %d)", path, size)
        return web.json_response(fileInfo)

    except Exception as e:
        logger.error("failed to upload file error: %s", str(e))
        return web.Response(
            body="Failed to upload, error: {}".format(str(e)), status=404
        )


cors = aiohttp_cors.setup(
    app,
    defaults={
        "*": aiohttp_cors.ResourceOptions(
            allow_credentials=True, expose_headers="*", allow_headers="*"
        )
    },
)

# app.router.add_post('/upload/{urlid}', upload_file)
cors.add(app.router.add_route("POST", "/upload/{urlid}", upload_file))


async def download_file(request):
    """Download file."""
    # origin = request.headers.get(hdrs.ORIGIN)
    # if origin is None:
    #     # Terminate CORS according to CORS 6.2.1.
    #     raise web.HTTPForbidden(
    #         text="CORS preflight request failed: "
    #              "origin header is not specified in the request")
    urlid = request.match_info["urlid"]  # Could be a HUGE file
    name = request.match_info["name"]
    if urlid not in generatedUrls:
        raise web.HTTPForbidden(text="Invalid URL")
    fileInfo = generatedUrls[urlid]
    if fileInfo.get("password", False):
        password = request.match_info.get("password")
        if password != fileInfo["password"]:
            raise web.HTTPForbidden(text="Incorrect password for accessing this file.")
    headers = fileInfo.get("headers")
    default_headers = {}
    if fileInfo["type"] == "dir":
        dirname = os.path.dirname(name)
        # list the folder
        if dirname == "" or dirname is None:
            if name != fileInfo["name"]:
                raise web.HTTPForbidden(text="File name does not match server record!")
            folder_path = fileInfo["path"]
            if not os.path.exists(folder_path):
                return web.Response(
                    body="Folder <{folder_path}> does not exist".format(
                        folder_path=folder_path
                    ),
                    status=404,
                )
            else:
                file_list = scandir(folder_path, "file", False)
                headers = headers or {
                    "Content-Disposition": 'inline; filename="{filename}"'.format(
                        filename=name
                    )
                }
                headers.update(default_headers)
                return web.json_response(file_list, headers=headers)
        # list the subfolder or get a file in the folder
        else:
            file_path = os.path.join(fileInfo["path"], os.sep.join(name.split("/")[1:]))
            if not os.path.exists(file_path):
                return web.Response(
                    body="File <{file_path}> does not exist".format(
                        file_path=file_path
                    ),
                    status=404,
                )
            if os.path.isdir(file_path):
                _, folder_name = os.path.split(file_path)
                file_list = scandir(file_path, "file", False)
                headers = headers or {
                    "Content-Disposition": 'inline; filename="{filename}"'.format(
                        filename=folder_name
                    )
                }
                headers.update(default_headers)
                return web.json_response(file_list, headers=headers)
            else:
                _, file_name = os.path.split(file_path)
                mime_type = (
                    MimeTypes().guess_type(file_name)[0] or "application/octet-stream"
                )
                file_size = os.path.getsize(file_path)
                headers = headers or {
                    "Content-Disposition": 'inline; filename="{filename}"'.format(
                        filename=file_name
                    ),
                    "Content-Type": mime_type,
                    "Content-Length": str(file_size),
                }
                headers.update(default_headers)
                return web.Response(
                    body=file_sender(file_path=file_path), headers=headers
                )
    elif fileInfo["type"] == "file":
        file_path = fileInfo["path"]
        if name != fileInfo["name"]:
            raise web.HTTPForbidden(text="File name does not match server record!")
        file_name = fileInfo["name"]
        if not os.path.exists(file_path):
            return web.Response(
                body="File <{file_name}> does not exist".format(file_name=file_path),
                status=404,
            )
        mime_type = MimeTypes().guess_type(file_name)[0] or "application/octet-stream"
        file_size = os.path.getsize(file_path)
        headers = headers or {
            "Content-Disposition": 'inline; filename="{filename}"'.format(
                filename=file_name
            ),
            "Content-Type": mime_type,
            "Content-Length": str(file_size),
        }
        headers.update(default_headers)
        return web.Response(body=file_sender(file_path=file_path), headers=headers)
    else:
        raise web.HTTPForbidden(text="Unsupported file type: " + fileInfo["type"])


cors.add(app.router.add_get("/file/{urlid}/{name:.+}", download_file))
cors.add(app.router.add_get("/file/{urlid}@{password}/{name:.+}", download_file))


@sio.on("get_file_url", namespace=NAME_SPACE)
async def on_get_file_url(sid, kwargs):
    """Return file url."""
    logger.info("generating file url: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered"}

    path = os.path.abspath(os.path.expanduser(kwargs["path"]))
    if not os.path.exists(path):
        return {"success": False, "error": "file does not exist."}
    fileInfo = {"path": path}
    if os.path.isdir(path):
        fileInfo["type"] = "dir"
    else:
        fileInfo["type"] = "file"
    if kwargs.get("headers"):
        fileInfo["headers"] = kwargs["headers"]
    _, name = os.path.split(path)
    fileInfo["name"] = name
    if path in generatedUrlFiles:
        return {"success": True, "url": generatedUrlFiles[path]}
    else:
        urlid = str(uuid.uuid4())
        generatedUrls[urlid] = fileInfo
        base_url = kwargs.get("base_url", registered_sessions[sid]["base_url"])
        if kwargs.get("password"):
            fileInfo["password"] = kwargs["password"]
            generatedUrlFiles[path] = "{}/file/{}@{}/{}".format(
                base_url, urlid, fileInfo["password"], name
            )
        else:
            generatedUrlFiles[path] = "{}/file/{}/{}".format(base_url, urlid, name)
        return {"success": True, "url": generatedUrlFiles[path]}


@sio.on("get_file_path", namespace=NAME_SPACE)
async def on_get_file_path(sid, kwargs):
    """Return file path."""
    logger.info("generating file url: %s", kwargs)
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered"}

    url = kwargs["url"]
    urlid = urlparse(url).path.replace("/file/", "")
    if urlid in generatedUrls:
        fileInfo = generatedUrls[urlid]
        return {"success": True, "path": fileInfo["path"]}
    else:
        return {"success": False, "error": "url not found."}


@sio.on("get_engine_status", namespace=NAME_SPACE)
async def on_get_engine_status(sid, kwargs):
    """Return engine status."""
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered."}
    current_process = psutil.Process()
    children = current_process.children(recursive=True)
    pid_dict = {}
    for i in plugins:
        p = plugins[i]
        pid_dict[p["process_id"]] = p
    procs = []
    for proc in children:
        if proc.is_running() and proc.status() != psutil.STATUS_ZOMBIE:
            if proc.pid in pid_dict:
                procs.append({"name": pid_dict[proc.pid]["name"], "pid": proc.pid})
            else:
                procs.append({"name": proc.name(), "pid": proc.pid})
    return {
        "success": True,
        "plugin_num": len(plugins),
        "plugin_processes": procs,
        "engine_process": current_process.pid,
    }


@sio.on("kill_plugin_process", namespace=NAME_SPACE)
async def on_kill_plugin_process(sid, kwargs):
    """Kill plugin process."""
    if sid not in registered_sessions:
        logger.debug("client %s is not registered.", sid)
        return {"success": False, "error": "client has not been registered."}
    if "all" not in kwargs:
        return {
            "success": False,
            "error": 'You must provide the pid of the plugin process or "all=true".',
        }
    if kwargs["all"]:
        logger.info("Killing all the plugins...")
        await killAllPlugins(sid)
        return {"success": True}
    else:
        try:
            print("Killing plugin process (pid=" + str(kwargs["pid"]) + ")...")
            killProcess(int(kwargs["pid"]))
            return {"success": True}
        except Exception:
            return {
                "success": False,
                "error": "Failed to kill plugin process: #" + str(kwargs["pid"]),
            }


@sio.on("message", namespace=NAME_SPACE)
async def on_message(sid, kwargs):
    """Receive message."""
    logger.info("message recieved: %s", kwargs)


@sio.on("disconnect", namespace=NAME_SPACE)
async def disconnect(sid):
    """Disconnect client."""
    disconnectClientSession(sid)
    disconnectPlugin(sid)
    logger.info("disconnect %s", sid)


def launch_plugin(
    stop_callback,
    logging_callback,
    pid,
    pname,
    tag,
    env,
    requirements,
    args,
    work_dir,
    abort,
    name,
    plugin_env,
):
    """Launch plugin."""
    if abort.is_set():
        logger.info("plugin aborting...")
        logging_callback("plugin aborting...")
        return False
    env_name = None
    try:
        repos = parseRepos(requirements, work_dir)
        logging_callback(2, type="progress")
        for k, r in enumerate(repos):
            try:
                print("Cloning repo " + r["url"] + " to " + r["repo_dir"])
                logging_callback("Cloning repo " + r["url"] + " to " + r["repo_dir"])
                if os.path.exists(r["repo_dir"]):
                    assert os.path.isdir(r["repo_dir"])
                    cmd = "git pull --all"
                    runCmd(cmd.split(" "), cwd=r["repo_dir"], plugin_id=pid)
                else:
                    cmd = (
                        "git clone --progress --depth=1 "
                        + r["url"]
                        + " "
                        + r["repo_dir"]
                    )
                    runCmd(cmd.split(" "), cwd=work_dir, plugin_id=pid)
                logging_callback(k * 5, type="progress")
            except Exception as ex:
                logging_callback(
                    "Failed to obtain the git repo: " + str(ex), type="error"
                )

        default_env_name = "{}-{}".format(pname, tag) if tag != "" else pname
        default_env_name = default_env_name.replace(" ", "_")
        env_name, envs, is_py2 = parseEnv(env, work_dir, default_env_name)
        default_requirements = (
            default_requirements_py2 if is_py2 else default_requirements_py3
        )
        requirements_cmd = parseRequirements(
            requirements, default_requirements, work_dir
        )

        if envs is not None and len(envs) > 0:
            for env in envs:
                print("Running env command: " + env)
                logger.info("running env command: %s", env)
                if env not in cmd_history:
                    logging_callback("running env command: {}".format(env))
                    process = subprocess.Popen(
                        env.split(),
                        shell=False,
                        env=plugin_env,
                        cwd=work_dir,
                        stderr=subprocess.PIPE,
                    )
                    setPluginPID(pid, process.pid)
                    ret = process.wait()
                    if ret == 0:
                        cmd_history.append(env)
                        logging_callback("env command executed successfully.")

                    _, errors = process.communicate()
                    if errors is not None:
                        logging_callback(str(errors, "utf-8"), type="error")

                    logging_callback(30, type="progress")
                else:
                    logger.debug("skip command: %s", env)
                    logging_callback("skip env command: " + env)

                if abort.is_set():
                    logger.info("plugin aborting...")
                    return False

        if opt.freeze:
            print(
                "WARNING: blocked pip command: \n{}\n"
                "You may want to run it yourself.".format(requirements_cmd)
            )
            logger.warning(
                "pip command is blocked due to `--freeze` mode: %s", requirements_cmd
            )
            requirements_cmd = None

        if not opt.freeze and CONDA_AVAILABLE:
            if env_name is not None:
                requirements_cmd = conda_activate.format(
                    env_name + " && " + requirements_cmd
                )

        logger.info("Running requirements command: %s", requirements_cmd)
        print("Running requirements command: " + requirements_cmd)
        if requirements_cmd is not None and requirements_cmd not in cmd_history:
            process = subprocess.Popen(
                requirements_cmd,
                shell=True,
                env=plugin_env,
                cwd=work_dir,
                stderr=subprocess.PIPE,
            )
            logging_callback(
                "Running requirements subprocess(pid={}): {}".format(
                    process.pid, requirements_cmd
                )
            )
            setPluginPID(pid, process.pid)
            ret = process.wait()
            _, errors = process.communicate()
            if ret != 0:
                logging_callback(
                    "Failed to run requirements command: {}".format(requirements_cmd),
                    type="error",
                )
                if errors is not None:
                    logging_callback(str(errors, "utf-8"), type="error")
                git_cmd = ""
                if shutil.which("git") is None:
                    git_cmd += " git"
                if shutil.which("pip") is None:
                    git_cmd += " pip"
                if git_cmd != "":
                    logger.info("pip command failed, trying to install git and pip...")
                    # try to install git and pip
                    git_cmd = "conda install -y" + git_cmd
                    process = subprocess.Popen(
                        git_cmd.split(), shell=False, env=plugin_env, cwd=work_dir
                    )
                    setPluginPID(pid, process.pid)
                    ret = process.wait()
                    if ret != 0:
                        logging_callback(
                            "Failed to install git/pip and dependencies "
                            "with exit code: " + str(ret),
                            type="error",
                        )
                        raise Exception(
                            "Failed to install git/pip and dependencies "
                            "with exit code: " + str(ret)
                        )
                    else:
                        process = subprocess.Popen(
                            requirements_cmd, shell=True, env=plugin_env, cwd=work_dir
                        )
                        setPluginPID(pid, process.pid)
                        ret = process.wait()
                        if ret != 0:
                            logging_callback(
                                "Failed to install dependencies with exit code: "
                                + str(ret),
                                type="error",
                            )
                            raise Exception(
                                "Failed to install dependencies with exit code: "
                                + str(ret)
                            )
            else:
                cmd_history.append(requirements_cmd)
                logging_callback("Requirements command executed successfully.")
            logging_callback(70, type="progress")
        else:
            logger.debug("skip command: %s", requirements_cmd)
    except Exception as e:
        # await sio.emit(
        #     'message_from_plugin_'+pid,
        #     {"type": "executeFailure", "error": "failed to install requirements."})
        logger.error(
            "Failed to setup plugin virtual environment or its requirements: %s", str(e)
        )
        logging_callback(
            "Failed to setup plugin virual environment or its requirements: " + str(e),
            type="error",
        )
        abort.set()

    if abort.is_set():
        logger.info("Plugin aborting...")
        logging_callback("Plugin aborting...")
        return False
    # env = os.environ.copy()
    if env_name is not None:
        args = conda_activate.format(env_name + " && " + args)
    if type(args) is str:
        args = args.split()
    if not args:
        args = []
    # Convert them all to strings
    args = [str(x) for x in args if str(x) != ""]
    logger.info("%s task started.", name)

    args = " ".join(args)
    logger.info("Task subprocess args: %s", args)

    # set system/version dependent "start_new_session" analogs
    # https://docs.python.org/2/library/subprocess.html#converting-argument-sequence
    kwargs = {}
    if sys.platform != "win32":
        kwargs.update(preexec_fn=os.setsid)
    logging_callback(100, type="progress")
    try:
        process = subprocess.Popen(
            args,
            bufsize=1,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            shell=True,
            env=plugin_env,
            cwd=work_dir,
            **kwargs
        )
        logging_callback("running subprocess(pid={}) with {}".format(process.pid, args))
        setPluginPID(pid, process.pid)
        # Poll process for new output until finished
        stdfn = sys.stdout.fileno()

        logging_callback(0, type="progress")

        while True:
            out = process.stdout.read(1)
            if out == "" and process.poll() is not None:
                break
            os.write(stdfn, out)
            sys.stdout.flush()
            if abort.is_set() or process.poll() is not None:
                break
            time.sleep(0)

        logger.info("Plugin aborting...")
        killProcess(process.pid)

        outputs, errors = process.communicate()
        if outputs is not None:
            outputs = str(outputs, "utf-8")
        if errors is not None:
            errors = str(errors, "utf-8")
        exitCode = process.returncode
    except Exception as e:
        outputs, errors = "", str(e)
        exitCode = 100
    finally:
        if exitCode == 0:
            logging_callback("plugin process exited with code {}".format(0))
            stop_callback(True, outputs)
            return True
        else:
            logging_callback(
                "plugin process exited with code {}".format(exitCode), type="error"
            )
            logger.info(
                "Error occured during terminating a process.\n"
                "command: %s\n exit code: %s\n",
                str(args),
                str(exitCode),
            )
            stop_callback(
                False,
                (errors or "")
                + "\nplugin process exited with code {}".format(exitCode),
            )
            return False


async def on_startup(app):
    """Start plugin engine."""
    print("ImJoy Python Plugin Engine (version {})".format(__version__))

    if opt.serve:
        print(
            "You can access your local ImJoy web app through "
            + opt.base_url
            + " , imjoy!"
        )
    else:
        print(
            "Please go to https://imjoy.io/#/app "
            "with your web browser (Chrome or FireFox)"
        )
    print("Connection Token: " + opt.token)
    sys.stdout.flush()


async def on_shutdown(app):
    """Shut down engine."""
    print("Shutting down...")
    logger.info("Shutting down the plugin engine...")
    stopped = threading.Event()

    def loop():  # executed in another thread
        for i in range(5):
            print("Exiting: " + str(5 - i), flush=True)
            time.sleep(0.5)
            if stopped.is_set():
                break
        print("Force shutting down now!", flush=True)
        logger.debug("Plugin engine is killed.")
        killProcess(os.getpid())
        # os._exit(1)

    t = threading.Thread(target=loop)
    t.daemon = True  # stop if the program exits
    t.start()

    print("Shutting down the plugins...", flush=True)
    # stopped.set()
    logger.info("Plugin engine exited.")
    try:
        os.remove(pid_file)
    except Exception:
        logger.info("Failed to remove the pid file.")


app.on_startup.append(on_startup)
app.on_shutdown.append(on_shutdown)


def main():
    """Run app."""
    try:
        web.run_app(app, host=opt.host, port=int(opt.port))
    except OSError as e:
        if e.errno in {48}:
            print(
                "ERROR: Failed to open port {}, "
                "please try to terminate the process which is using that port, "
                "or restart your computer.".format(opt.port)
            )


if __name__ == "__main__":
    main()



"""Provide a worker template."""
import argparse
import inspect
import logging
import math
import os
import sys
import threading
from functools import reduce
from types import ModuleType

from imjoySocketIO_client import LoggingNamespace, SocketIO, find_callback
from imjoyUtils import ReferenceStore, debounce, dotdict, setInterval

if sys.version_info >= (3, 0):
    import asyncio
    import janus
    from imjoyUtils3 import task_worker, FuturePromise

    PYTHON3 = True
else:
    from imjoyUtils import task_worker, Promise

    PYTHON3 = False

try:
    import queue
except ImportError:
    import Queue as queue

logging.basicConfig(stream=sys.stdout)
logger = logging.getLogger("plugin")
logger.setLevel(logging.INFO)
# import logging
# logging.basicConfig(level=logging.DEBUG)
ARRAY_CHUNK = 1000000

if "" not in sys.path:
    sys.path.insert(0, "")

imjoy_path = os.path.dirname(os.path.normpath(__file__))
if imjoy_path not in sys.path:
    sys.path.insert(0, imjoy_path)


def kill(proc_pid):
    """Kill process id."""
    import psutil

    process = psutil.Process(proc_pid)
    for proc in process.children(recursive=True):
        proc.kill()
    process.kill()


def ndarray(typedArray, shape, dtype):
    """Return a ndarray."""
    _dtype = type(typedArray)
    if dtype and dtype != _dtype:
        raise Exception(
            "dtype doesn't match the type of the array: " + _dtype + " != " + dtype
        )
    shape = shape or (len(typedArray),)
    return {
        "__jailed_type__": "ndarray",
        "__value__": typedArray,
        "__shape__": shape,
        "__dtype__": _dtype,
    }


api_utils = dotdict(
    ndarray=ndarray, kill=kill, debounce=debounce, setInterval=setInterval
)


class PluginConnection:
    """Represent a plugin connection."""

    def __init__(
        self,
        pid,
        secret,
        server,
        queue=None,
        loop=None,
        worker=None,
        namespace="/",
        work_dir=None,
        daemon=False,
        api=None,
    ):
        """Set up connection."""
        if work_dir is None or work_dir == "" or work_dir == ".":
            self.work_dir = os.getcwd()
        else:
            self.work_dir = work_dir
            if not os.path.exists(self.work_dir):
                os.makedirs(self.work_dir)
            os.chdir(self.work_dir)
        socketIO = SocketIO(server, Namespace=LoggingNamespace)
        self.socketIO = socketIO
        self._init = False
        self.secret = secret
        self.id = pid
        self.daemon = daemon

        def emit(msg):
            socketIO.emit("from_plugin_" + secret, msg)

        self.emit = emit

        self._local = {}
        _remote = dotdict()
        self._setLocalAPI(_remote)
        self._interface = {}
        self._plugin_interfaces = {}
        self._remote_set = False
        self._store = ReferenceStore()
        self._executed = False
        self.queue = queue
        self.loop = loop

        self._init = False
        sys.stdout.flush()
        socketIO.on("to_plugin_" + secret, self.sio_plugin_message)
        self.emit({"type": "initialized", "dedicatedThread": True})
        print('Plugin "{}" Initialized.'.format(pid))

        def on_disconnect():
            if not self.daemon:
                self.exit(1)

        socketIO.on("disconnect", on_disconnect)
        self.abort = threading.Event()
        self.worker = worker

    def wait_forever(self):
        """Wait forever."""
        if PYTHON3:
            self.sync_q = self.queue.sync_q
            fut = self.loop.run_in_executor(None, self.socketIO.wait)
            t = [
                self.worker(self, self.queue.async_q, logger, self.abort)
                for i in range(10)
            ]
            self.loop.run_until_complete(asyncio.gather(*t))
            self.loop.run_until_complete(fut)
        else:
            self.sync_q = queue.Queue()
            t = threading.Thread(target=self.socketIO.wait)
            t.daemon = True
            t.start()
            self.worker(self, self.sync_q, logger, self.abort)

    def default_exit(self):
        """Exit default."""
        logger.info("terminating plugin: " + self.id)
        self.abort.set()
        os._exit(0)

    def exit(self, code):
        """Exit."""
        if "exit" in self._interface:
            try:
                self._interface["exit"]()
            except Exception as e:
                logger.error("Error when exiting: %s", e)
                sys.exit(1)
            else:
                logger.info("terminating plugin")
                sys.exit(code)
        else:
            sys.exit(0)

    def _encode(self, aObject):
        if aObject is None:
            return aObject
        if type(aObject) is tuple:
            aObject = list(aObject)
        isarray = type(aObject) is list
        bObject = [] if isarray else {}
        # skip if already encoded
        if (
            type(aObject) is dict
            and "__jailed_type__" in aObject
            and "__value__" in aObject
        ):
            return aObject

        # encode interfaces
        if (
            type(aObject) is dict
            and "__id__" in aObject
            and "__jailed_type__" in aObject
            and aObject["__jailed_type__"] == "plugin_api"
        ):
            encoded_interface = {}
            for k in aObject.keys():
                v = aObject[k]
                if callable(v):
                    bObject[k] = {
                        "__jailed_type__": "plugin_interface",
                        "__plugin_id__": aObject["__id__"],
                        "__value__": k,
                        "num": None,
                    }
                    encoded_interface[k] = v
            self._plugin_interfaces[aObject["__id__"]] = encoded_interface
            return bObject

        keys = range(len(aObject)) if isarray else aObject.keys()
        for k in keys:
            v = aObject[k]
            try:
                basestring
            except NameError:
                basestring = str
            if callable(v):
                interfaceFuncName = None
                for name in self._interface:
                    if self._interface[name] == v:
                        interfaceFuncName = name
                        break
                if interfaceFuncName is None:
                    cid = self._store.put(v)
                    vObj = {"__jailed_type__": "callback", "__value__": "f", "num": cid}
                else:
                    vObj = {
                        "__jailed_type__": "interface",
                        "__value__": interfaceFuncName,
                    }

            # send objects supported by structure clone algorithm
            # https://developer.mozilla.org/en-US/docs/Web/API/Web_Workers_API/Structured_clone_algorithm
            # if (
            #   v !== Object(v) ||
            #   v instanceof Boolean ||
            #   v instanceof String ||
            #   v instanceof Date ||
            #   v instanceof RegExp ||
            #   v instanceof Blob ||
            #   v instanceof File ||
            #   v instanceof FileList ||
            #   v instanceof ArrayBuffer ||
            #   v instanceof ArrayBufferView ||
            #   v instanceof ImageData
            # ) {
            # }
            elif "np" in self._local and isinstance(
                v, (self._local["np"].ndarray, self._local["np"].generic)
            ):
                vb = bytearray(v.tobytes())
                if len(vb) > ARRAY_CHUNK:
                    vl = int(math.ceil(1.0 * len(vb) / ARRAY_CHUNK))
                    v_bytes = []
                    for i in range(vl):
                        v_bytes.append(vb[i * ARRAY_CHUNK : (i + 1) * ARRAY_CHUNK])
                else:
                    v_bytes = vb
                vObj = {
                    "__jailed_type__": "ndarray",
                    "__value__": v_bytes,
                    "__shape__": v.shape,
                    "__dtype__": str(v.dtype),
                }
            elif type(v) is dict or type(v) is list:
                vObj = self._encode(v)
            elif not isinstance(v, basestring) and type(v) is bytes:
                vObj = v.decode()  # covert python3 bytes to str
            elif isinstance(v, Exception):
                vObj = {"__jailed_type__": "error", "__value__": str(v)}
            else:
                vObj = {"__jailed_type__": "argument", "__value__": v}

            if isarray:
                bObject.append(vObj)
            else:
                bObject[k] = vObj

        return bObject

    def _decode(self, aObject, callbackId, withPromise):
        if aObject is None:
            return aObject
        if "__jailed_type__" in aObject and "__value__" in aObject:
            if aObject["__jailed_type__"] == "callback":
                bObject = self._genRemoteCallback(
                    callbackId, aObject["num"], withPromise
                )
            elif aObject["__jailed_type__"] == "interface":
                name = aObject["__value__"]
                if name in self._remote:
                    bObject = self._remote[name]
                else:
                    bObject = self._genRemoteMethod(name)
            elif aObject["__jailed_type__"] == "plugin_interface":
                bObject = self._genRemoteMethod(
                    aObject["__value__"], aObject["__plugin_id__"]
                )
            elif aObject["__jailed_type__"] == "ndarray":
                # create build array/tensor if used in the plugin
                try:
                    np = self._local["np"]
                    if isinstance(aObject["__value__"], bytearray):
                        aObject["__value__"] = aObject["__value__"]
                    elif isinstance(aObject["__value__"], list) or isinstance(
                        aObject["__value__"], tuple
                    ):
                        aObject["__value__"] = reduce(
                            (lambda x, y: x + y), aObject["__value__"]
                        )
                    else:
                        raise Exception(
                            "Unsupported data type: ",
                            type(aObject["__value__"]),
                            aObject["__value__"],
                        )
                    bObject = np.frombuffer(
                        aObject["__value__"], dtype=aObject["__dtype__"]
                    ).reshape(tuple(aObject["__shape__"]))
                except Exception as e:
                    logger.debug("Error in converting: %s", e)
                    bObject = aObject
                    raise e
            elif aObject["__jailed_type__"] == "error":
                bObject = Exception(aObject["__value__"])
            elif aObject["__jailed_type__"] == "argument":
                bObject = aObject["__value__"]
            else:
                bObject = aObject["__value__"]
            return bObject
        else:
            if isinstance(aObject, tuple):
                aObject = list(aObject)
            isarray = isinstance(aObject, list)
            bObject = [] if isarray else dotdict()
            keys = range(len(aObject)) if isarray else aObject.keys()
            for k in keys:
                if isarray or k in aObject:
                    v = aObject[k]
                    if isinstance(v, dict) or isinstance(v, list):
                        if isarray:
                            bObject.append(self._decode(v, callbackId, withPromise))
                        else:
                            bObject[k] = self._decode(v, callbackId, withPromise)
            return bObject

    def _wrap(self, args):
        wrapped = self._encode(args)
        result = {"args": wrapped}
        return result

    def _unwrap(self, args, withPromise):
        if "callbackId" not in args:
            args["callbackId"] = None
        # wraps each callback so that the only one could be called
        result = self._decode(args["args"], args["callbackId"], withPromise)
        return result

    def setInterface(self, api):
        """Set interface."""
        if isinstance(api, dict):
            api = {a: api[a] for a in api.keys() if not a.startswith("_")}
        elif inspect.isclass(type(api)):
            api = {a: getattr(api, a) for a in dir(api) if not a.startswith("_")}
        else:
            raise Exception("unsupported api export")
        if "exit" in api:
            ext = api["exit"]

            def exit_wrapper():
                try:
                    ext()
                finally:
                    self.default_exit()

            api["exit"] = exit_wrapper
        else:
            api["exit"] = self.default_exit
        self._interface = api
        self._sendInterface()

    def _sendInterface(self):
        names = []
        for name in self._interface:
            if callable(self._interface[name]):
                names.append({"name": name, "data": None})
            else:
                data = self._interface[name]
                if data is not None and isinstance(data, dict):
                    data2 = {}
                    for k in data:
                        if callable(data[k]):
                            data2[k] = "**@@FUNCTION@@**:" + k
                        else:
                            data2[k] = data[k]
                    names.append({"name": name, "data": data2})
                elif type(data) in [str, int, float, bool]:
                    names.append({"name": name, "data": data})
        self.emit({"type": "setInterface", "api": names})

    def _genRemoteMethod(self, name, plugin_id=None):
        def remoteMethod(*arguments, **kwargs):
            # wrap keywords to a dictionary and pass to the first argument
            if len(arguments) == 0 and len(kwargs) > 0:
                arguments = [kwargs]

            def p(resolve, reject):
                resolve.__jailed_pairs__ = reject
                reject.__jailed_pairs__ = resolve
                call_func = {
                    "type": "method",
                    "name": name,
                    "pid": plugin_id,
                    "args": self._wrap(arguments),
                    "promise": self._wrap([resolve, reject]),
                }
                self.emit(call_func)

            if PYTHON3:
                return FuturePromise(p, self.loop)
            else:
                return Promise(p)

        remoteMethod.__remote_method = True
        return remoteMethod

    def _genRemoteCallback(self, id, argNum, withPromise):
        if withPromise:

            def remoteCallback(*arguments, **kwargs):
                # wrap keywords to a dictionary and pass to the first argument
                if len(arguments) == 0 and len(kwargs) > 0:
                    arguments = [kwargs]

                def p(resolve, reject):
                    resolve.__jailed_pairs__ = reject
                    reject.__jailed_pairs__ = resolve
                    self.emit(
                        {
                            "type": "callback",
                            "id": id,
                            "num": argNum,
                            # 'pid'  : self.id,
                            "args": self._wrap(arguments),
                            "promise": self._wrap([resolve, reject]),
                        }
                    )

                if PYTHON3:
                    return FuturePromise(p, self.loop)
                else:
                    return Promise(p)

        else:

            def remoteCallback(*arguments, **kwargs):
                # wrap keywords to a dictionary and pass to the first argument
                if len(arguments) == 0 and len(kwargs) > 0:
                    arguments = [kwargs]
                ret = self.emit(
                    {
                        "type": "callback",
                        "id": id,
                        "num": argNum,
                        # 'pid'  : self.id,
                        "args": self._wrap(arguments),
                    }
                )
                return ret

        return remoteCallback

    def _setRemote(self, api):
        _remote = dotdict()
        for i in range(len(api)):
            if isinstance(api[i], dict) and "name" in api[i]:
                name = api[i]["name"]
                data = api[i].get("data", None)
                if data is not None:
                    if isinstance(data, dict):
                        data2 = dotdict()
                        for key in data:
                            if key in data:
                                if data[key] == "**@@FUNCTION@@**:" + key:
                                    data2[key] = self._genRemoteMethod(name + "." + key)
                                else:
                                    data2[key] = data[key]
                        _remote[name] = data2
                    else:
                        _remote[name] = data
                else:
                    _remote[name] = self._genRemoteMethod(name)

        self._setLocalAPI(_remote)
        return _remote

    def _setLocalAPI(self, _remote):
        _remote["export"] = self.setInterface
        _remote["utils"] = api_utils
        _remote["WORK_DIR"] = self.work_dir

        self._local["api"] = _remote

        # make a fake module with api
        m = ModuleType("imjoy")
        sys.modules[m.__name__] = m
        m.__file__ = m.__name__ + ".py"
        m.api = _remote

    def sio_plugin_message(self, *args):
        """Handle plugin message."""
        data = args[0]
        if data["type"] == "import":
            self.emit({"type": "importSuccess", "url": data["url"]})
        elif data["type"] == "disconnect":
            self.abort.set()
            callback, args = find_callback(args)
            try:
                if "exit" in self._interface and callable(self._interface["exit"]):
                    self._interface["exit"]()
            except Exception as e:
                logger.error("Error when exiting: %s", e)
            if callback:
                callback(*args)
        elif data["type"] == "execute":
            if not self._executed:
                self.sync_q.put(data)
            else:
                logger.debug("skip execution.")
                self.emit({"type": "executeSuccess"})
        elif data["type"] == "message":
            d = data["data"]
            self.sync_q.put(d)
            logger.debug("added task to the queue")
        sys.stdout.flush()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--id", type=str, required=True, help="plugin id")
    parser.add_argument("--secret", type=str, required=True, help="plugin secret")
    parser.add_argument("--namespace", type=str, default="/", help="socketio namespace")
    parser.add_argument(
        "--work_dir", type=str, default=".", help="plugin working directory"
    )
    parser.add_argument(
        "--server", type=str, default="http://127.0.0.1:9527", help="socketio server"
    )
    parser.add_argument("--daemon", action="store_true", help="daemon mode")
    parser.add_argument("--debug", action="store_true", help="debug mode")

    opt = parser.parse_args()
    if opt.debug:
        logger.setLevel(logging.DEBUG)

    if PYTHON3:
        loop = asyncio.get_event_loop()
        q = janus.Queue(loop=loop)
    else:
        loop = None
        q = None

    pc = PluginConnection(
        opt.id,
        opt.secret,
        server=opt.server,
        work_dir=opt.work_dir,
        queue=q,
        loop=loop,
        worker=task_worker,
    )
    pc.wait_forever()



import six

try:
    from logging import NullHandler
except ImportError:  # Python 2.6
    from logging import Handler

    class NullHandler(Handler):
        def emit(self, record):
            pass


try:
    from urllib import urlencode as format_query
except ImportError:
    from urllib.parse import urlencode as format_query
try:
    from urlparse import urlparse as parse_url
except ImportError:
    from urllib.parse import urlparse as parse_url
try:
    memoryview = memoryview
except NameError:
    memoryview = buffer


def get_int(*args):
    try:
        return int(get_character(*args))
    except ValueError:
        return ord(get_character(*args))


def get_character(x, index):
    return chr(get_byte(x, index))


def get_byte(x, index):
    return six.indexbytes(x, index)


def encode_string(x):
    return x.encode("utf-8")


def decode_string(x):
    return x.decode("utf-8")



import requests
import six
import socket
import ssl
import sys
import threading
import time
import websocket

from .exceptions import ConnectionError, TimeoutError
from .parsers import (
    encode_engineIO_content,
    decode_engineIO_content,
    format_packet_text,
    parse_packet_text,
)
from .symmetries import format_query, memoryview, parse_url

# From https://github.com/invisibleroads/socketIO-client/pull/139#issuecomment-265124962
try:
    from ssl import SSLError
except ImportError:
    # dummy class of SSLError for ssl none-support environment.
    class SSLError(Exception):
        pass


if not hasattr(websocket, "create_connection"):
    sys.exit(
        """\
An incompatible websocket library is conflicting with the one we need.
You can remove the incompatible library and install the correct one
by running the following commands:

yes | pip uninstall websocket websocket-client
pip install -U websocket-client"""
    )


ENGINEIO_PROTOCOL = 3
TRANSPORTS = "xhr-polling", "websocket"


class AbstractTransport(object):
    def __init__(self, http_session, is_secure, url, engineIO_session=None):
        self.http_session = http_session
        self.is_secure = is_secure
        self.url = url
        self.engineIO_session = engineIO_session

    def recv_packet(self):
        pass

    def send_packet(self, engineIO_packet_type, engineIO_packet_data=""):
        pass

    def set_timeout(self, seconds=None):
        pass

    def close(self):
        pass


class XHR_PollingTransport(AbstractTransport):
    def __init__(self, http_session, is_secure, url, engineIO_session=None):
        super(XHR_PollingTransport, self).__init__(
            http_session, is_secure, url, engineIO_session
        )
        self._params = {"EIO": ENGINEIO_PROTOCOL, "transport": "polling"}
        if engineIO_session:
            self._request_index = 1
            self._kw_get = dict(timeout=engineIO_session.ping_timeout)
            self._kw_post = dict(
                timeout=engineIO_session.ping_timeout,
                headers={"content-type": "application/octet-stream"},
            )
            self._params["sid"] = engineIO_session.id
        else:
            self._request_index = 0
            self._kw_get = {}
            self._kw_post = {}
        http_scheme = "https" if is_secure else "http"
        self._http_url = "%s://%s/" % (http_scheme, url)
        self._request_index_lock = threading.Lock()
        self._send_packet_lock = threading.Lock()

    def recv_packet(self):
        params = dict(self._params)
        params["t"] = self._get_timestamp()
        response = get_response(
            self.http_session.get, self._http_url, params=params, **self._kw_get
        )
        for engineIO_packet in decode_engineIO_content(response.content):
            engineIO_packet_type, engineIO_packet_data = engineIO_packet
            yield engineIO_packet_type, engineIO_packet_data

    def send_packet(self, engineIO_packet_type, engineIO_packet_data=""):
        with self._send_packet_lock:
            params = dict(self._params)
            params["t"] = self._get_timestamp()
            data = encode_engineIO_content(
                [(engineIO_packet_type, engineIO_packet_data)]
            )
            response = get_response(
                self.http_session.post,
                self._http_url,
                params=params,
                data=memoryview(data),
                **self._kw_post
            )
            assert response.content == b"ok"

    def _get_timestamp(self):
        with self._request_index_lock:
            timestamp = "%s-%s" % (int(time.time() * 1000), self._request_index)
            self._request_index += 1
        return timestamp


class WebsocketTransport(AbstractTransport):
    def __init__(self, http_session, is_secure, url, engineIO_session=None):
        super(WebsocketTransport, self).__init__(
            http_session, is_secure, url, engineIO_session
        )
        params = dict(
            http_session.params, **{"EIO": ENGINEIO_PROTOCOL, "transport": "websocket"}
        )
        request = http_session.prepare_request(requests.Request("GET", url))
        kw = {"header": ["%s: %s" % x for x in request.headers.items()]}
        if engineIO_session:
            params["sid"] = engineIO_session.id
            kw["timeout"] = self._timeout = engineIO_session.ping_timeout
        ws_url = "%s://%s/?%s" % (
            "wss" if is_secure else "ws",
            url,
            format_query(params),
        )
        http_scheme = "https" if is_secure else "http"
        if http_scheme in http_session.proxies:  # Use the correct proxy
            proxy_url_pack = parse_url(http_session.proxies[http_scheme])
            kw["http_proxy_host"] = proxy_url_pack.hostname
            kw["http_proxy_port"] = proxy_url_pack.port
            if proxy_url_pack.username:
                kw["http_proxy_auth"] = (
                    proxy_url_pack.username,
                    proxy_url_pack.password,
                )
        if http_session.verify:
            if http_session.cert:  # Specify certificate path on disk
                if isinstance(http_session.cert, basestring):
                    kw["ca_certs"] = http_session.cert
                else:
                    kw["ca_certs"] = http_session.cert[0]
        else:  # Do not verify the SSL certificate
            kw["sslopt"] = {"cert_reqs": ssl.CERT_NONE}
        try:
            self._connection = websocket.create_connection(ws_url, **kw)
        except Exception as e:
            raise ConnectionError(e)

    def recv_packet(self):
        try:
            packet_text = self._connection.recv()
        except websocket.WebSocketTimeoutException as e:
            raise TimeoutError("recv timed out (%s)" % e)
        except SSLError as e:
            raise ConnectionError("recv disconnected by SSL (%s)" % e)
        except websocket.WebSocketConnectionClosedException as e:
            raise ConnectionError("recv disconnected (%s)" % e)
        except socket.error as e:
            raise ConnectionError("recv disconnected (%s)" % e)
        if not isinstance(packet_text, six.binary_type):
            packet_text = six.b(packet_text)
        engineIO_packet_type, engineIO_packet_data = parse_packet_text(packet_text)
        yield engineIO_packet_type, engineIO_packet_data

    def send_packet(self, engineIO_packet_type, engineIO_packet_data=""):
        packet = format_packet_text(engineIO_packet_type, engineIO_packet_data)
        try:
            self._connection.send(packet)
        except websocket.WebSocketTimeoutException as e:
            raise TimeoutError("send timed out (%s)" % e)
        except socket.error as e:
            raise ConnectionError("send disconnected (%s)" % e)
        except websocket.WebSocketConnectionClosedException as e:
            raise ConnectionError("send disconnected (%s)" % e)

    def set_timeout(self, seconds=None):
        self._connection.settimeout(seconds or self._timeout)

    def close(self):
        self._connection.close()


def get_response(request, *args, **kw):
    try:
        response = request(*args, stream=True, **kw)
    except requests.exceptions.Timeout as e:
        raise TimeoutError(e)
    except requests.exceptions.ConnectionError as e:
        raise ConnectionError(e)
    except requests.exceptions.SSLError as e:
        raise ConnectionError("could not negotiate SSL (%s)" % e)
    status_code = response.status_code
    if 200 != status_code:
        raise ConnectionError(
            "unexpected status code (%s %s)" % (status_code, response.text)
        )
    return response


def prepare_http_session(kw):
    http_session = requests.Session()
    http_session.headers.update(kw.get("headers", {}))
    http_session.auth = kw.get("auth")
    http_session.proxies.update(kw.get("proxies", {}))
    http_session.hooks.update(kw.get("hooks", {}))
    http_session.params.update(kw.get("params", {}))
    http_session.verify = kw.get("verify", True)
    http_session.cert = kw.get("cert")
    http_session.cookies.update(kw.get("cookies", {}))
    return http_session



"""Set up the ImJoy-Engine imjoy package."""
import os
import sys
import pathlib
import subprocess
from setuptools import setup, find_packages
import json

DESCRIPTION = (
    "ImJoy Plugin Engine for running Python plugins locally "
    "or remotely from ImJoy.io"
)

try:
    subprocess.call(["conda", "-V"])
    if os.name == "nt":
        # for fixing CondaHTTPError:
        # https://github.com/conda/conda/issues/6064#issuecomment-458389796
        process = subprocess.Popen(
            ["conda", "info", "--json", "-s"], stdout=subprocess.PIPE
        )
        cout, err = process.communicate()
        conda_prefix = json.loads(cout.decode("ascii"))["conda_prefix"]
        print("Found conda environment: " + conda_prefix)
        os.environ["PATH"] = (
            os.path.join(conda_prefix, "Library", "bin")
            + os.pathsep
            + os.environ["PATH"]
        )
except OSError:
    if sys.version_info > (3, 0):
        print(
            "WARNING: you are running ImJoy without conda, "
            "you may have problem with some plugins."
        )
    else:
        sys.exit(
            "Sorry, ImJoy Python Plugin Engine can only run within a conda environment "
            "or at least in Python 3."
        )

requirements = []
if sys.version_info > (3, 0):
    requirements = [
        "aiohttp",
        "aiohttp_cors",
        "python-socketio",
        "requests",
        "six",
        "websocket-client",
        "numpy",
        "janus",
        "pyyaml",
    ]
    print("Trying to install psutil with pip...")
    ret = subprocess.Popen(
        ["pip", "install", "psutil"], env=os.environ.copy(), shell=False
    ).wait()
    if ret != 0:
        print("Trying to install psutil with conda...")
        ret2 = subprocess.Popen(
            ["conda", "install", "-y", "psutil"], env=os.environ.copy()
        ).wait()
        if ret2 != 0:
            raise Exception(
                "Failed to install psutil, "
                "please try to setup an environment with gcc support."
            )


HERE = pathlib.Path(__file__).parent
README = (HERE / "README.md").read_text()
VERSION = json.loads((HERE / "imjoy" / "VERSION").read_text())["version"]

setup(
    name="imjoy",
    version=VERSION,
    description=DESCRIPTION,
    long_description=README,
    long_description_content_type="text/markdown",
    url="http://github.com/oeway/ImJoy-Engine",
    author="Wei OUYANG",
    author_email="oeway007@gmail.com",
    license="MIT",
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    zip_safe=False,
    entry_points={"console_scripts": ["imjoy = imjoy.__main__:main"]},
)
